-- The top module of Deep Learning on FPGA
-- by Toshio Iwata at digitalfilter.com 2019/02/13

library ieee ;
use ieee.std_logic_1164.all ;
use ieee.std_logic_arith.all ;
use ieee.std_logic_signed.all;

-------------------------------------------------------------------------------
-- Entity
-------------------------------------------------------------------------------
entity ssd_hw is
port(
  CLK : in std_logic;                            -- clock (100MHz)
  RST_N : in std_logic;                          -- reset (active low)
  WR_BUF_SEL : in std_logic_vector(1 downto 0);
  WR_BUF_SEL2 : in std_logic_vector(1 downto 0);
  WR_BUF_N : in std_logic;
  RD_BUF_SEL : in std_logic_vector(1 downto 0);
  RD_BUF_ADDR : in std_logic_vector(7 downto 0);
  SsdLineCount : in std_logic_vector(8 downto 0);
  SsdLineCount2 : in std_logic_vector(8 downto 0);
  SsdPixCount : in std_logic_vector(8 downto 0);
  SsdPixCount2 : in std_logic_vector(8 downto 0);
  OutChSel : in std_logic_vector(1 downto 0);
  CO9 : in std_logic_vector(3 downto 0);
  CO512 : in std_logic_vector(8 downto 0);
  CO512_dly2 : in std_logic_vector(8 downto 0);
--		BigLineN : in std_logic_vector(8 downto 0);
		DISPJUDGE : in std_logic;
  ReluWrN : out std_logic;
  ReluWrData : out std_logic_vector(2559 downto 0);
  ReluWrAddr : out std_logic_vector(5 downto 0);
  ReluWrSel : out std_logic_vector(1 downto 0);
  LayNextWrN : in std_logic;
  LayNextWrData : in std_logic_vector(2559 downto 0);
  LayNextWrAddr : in std_logic_vector(5 downto 0);
  LayNextWrSel : in std_logic_vector(1 downto 0);
		ReluSum : out std_logic_vector(11 downto 0) );
end ssd_hw;

-------------------------------------------------------------------------------
-- Architecture
-------------------------------------------------------------------------------
architecture RTL of ssd_hw is

component conv_layer9
port(
  CLK : in std_logic;                            -- clock (100MHz)
  RST_N : in std_logic;                          -- reset (active low)
  OutChSel : in std_logic_vector(1 downto 0);
  CO300 : in std_logic_vector(8 downto 0);
  CO9 : in std_logic_vector(3 downto 0);
  CO512 : in std_logic_vector(8 downto 0);
  CO512_dly2 : in std_logic_vector(8 downto 0);
  X_RD_DATA : in std_logic_vector(2559 downto 0); 
  ReluAll : out std_logic_vector(2559 downto 0); 
  ReluAllWrN : out std_logic; 
  ReluSum : out std_logic_vector(11 downto 0) ); 
end component;

component SRAM_64x2560
Port (
          WRCLK : In std_logic; 
          RDCLK : In std_logic;
          WR_N : In std_logic;
          WRADDR : In std_logic_vector(5 downto 0);
          RDADDR : In std_logic_vector(5 downto 0);
          WRDATA : In std_logic_vector(2559 downto 0);
          RDDATA : Out std_logic_vector(2559 downto 0) );
end component;

--component pooling_lay7
--Port (
--  CLK : in std_logic;                            -- clock (100MHz)
--  RST_N : in std_logic;                          -- reset (active low)
--  ReluAll : in std_logic_vector(1279 downto 0); 
--  WR_N : in std_logic; 
--  HCO75 : in std_logic_vector(6 downto 0); 
--  HCOEDGE : in std_logic; 
--  Poolout : Out std_logic_vector(1279 downto 0) );  
--end component;

-------------------------------------------------------------------------------
-- Signal
-------------------------------------------------------------------------------
signal BchData : std_logic_vector(6 downto 0);
signal GchData : std_logic_vector(6 downto 0);
signal RchData : std_logic_vector(6 downto 0);
signal Bch_oft, Bch : std_logic_vector(5 downto 0);
signal Gch_oft, Gch : std_logic_vector(6 downto 0);
signal Rch_oft, Rch : std_logic_vector(5 downto 0);

signal RAM_X1_WR_N_0, RAM_X1_WR_N_1, RAM_X1_WR_N_2, RAM_X1_WR_N_3 : std_logic;
signal RAM_X1_WR_ADDR : std_logic_vector(5 downto 0);
signal RAM_X1_RD_ADDR : std_logic_vector(5 downto 0);
signal RAM_X1_RD_DATA_0, RAM_X1_RD_DATA_1, RAM_X1_RD_DATA_2, RAM_X1_RD_DATA_3 : std_logic_vector(2559 downto 0);
signal RAM_X1_RD_DATA_2_padding : std_logic_vector(2559 downto 0);
signal RAM_X1_RD_DATA_1_padding : std_logic_vector(2559 downto 0);
signal RAM_X1_RD_DATA : std_logic_vector(2559 downto 0);
signal RAM_X1_RD_SEL : std_logic_vector(1 downto 0);

signal RAM_X2_WR_N_0, RAM_X2_WR_N_1, RAM_X2_WR_N_2, RAM_X2_WR_N_3 : std_logic;
signal RAM_X2_WR_ADDR : std_logic_vector(5 downto 0);
signal RAM_X2_RD_ADDR : std_logic_vector(5 downto 0);
signal RAM_X2_WR_DATA : std_logic_vector(2559 downto 0);
signal RAM_X2_RD_DATA_0, RAM_X2_RD_DATA_1, RAM_X2_RD_DATA_2, RAM_X2_RD_DATA_3 : std_logic_vector(2559 downto 0);
signal RAM_X2_RD_DATA : std_logic_vector(2559 downto 0);

signal ReluAll : std_logic_vector(2559 downto 0);
signal ReluAllWrN : std_logic;
signal ReluWrN_sig : std_logic;
signal ReluWrN_tmp : std_logic;
signal HCO150 : std_logic_vector(8 downto 0);
signal tmpout000, tmpout001 : std_logic_vector(4 downto 0);
signal HCO75 : std_logic_vector(6 downto 0);
signal Pooladdr : std_logic_vector(6 downto 0);
signal Poolout : std_logic_vector(2559 downto 0);
signal hco75is75, hco75is75_dly1, hco75is75_dly2, HCOEDGE : std_logic;

begin

---------------------------------------------------------
  ReluWrN <= ReluWrN_sig;
  ReluWrSel <= WR_BUF_SEL;
  ReluWrData <= RAM_X2_WR_DATA;
  ReluWrAddr <= RAM_X2_WR_ADDR;

---------------------------------------------------------
RAM_X1_0 : SRAM_64x2560 port map (
   WRCLK	=> CLK,
   RDCLK	=> CLK,
   WR_N		=> RAM_X1_WR_N_0, 
   WRADDR	=> RAM_X1_WR_ADDR, 
   RDADDR	=> RAM_X1_RD_ADDR, 
   WRDATA	=> LayNextWrData, -- RAM_X1_WR_DATA, 
   RDDATA	=> RAM_X1_RD_DATA_0);

-- RAM_X1_WR_N_0 <= '0' when WR_BUF_N = '0' and WR_BUF_SEL = "00" else '1';
RAM_X1_WR_N_0 <= '0' when LayNextWrN = '0' and LayNextWrSel = "00" else '1';

---------------------------------------------------------
RAM_X1_1 : SRAM_64x2560 port map (
   WRCLK	=> CLK,
   RDCLK	=> CLK,
   WR_N		=> RAM_X1_WR_N_1, 
   WRADDR	=> RAM_X1_WR_ADDR, 
   RDADDR	=> RAM_X1_RD_ADDR, 
   WRDATA	=> LayNextWrData, -- RAM_X1_WR_DATA, 
   RDDATA	=> RAM_X1_RD_DATA_1);

-- RAM_X1_WR_N_1 <= '0' when WR_BUF_N = '0' and WR_BUF_SEL = "01" else '1';
RAM_X1_WR_N_1 <= '0' when LayNextWrN = '0' and LayNextWrSel = "01" else '1';

---------------------------------------------------------
RAM_X1_2 : SRAM_64x2560 port map (
   WRCLK	=> CLK,
   RDCLK	=> CLK,
   WR_N		=> RAM_X1_WR_N_2, 
   WRADDR	=> RAM_X1_WR_ADDR, 
   RDADDR	=> RAM_X1_RD_ADDR, 
   WRDATA	=> LayNextWrData, -- RAM_X1_WR_DATA, 
   RDDATA	=> RAM_X1_RD_DATA_2);

-- RAM_X1_WR_N_2 <= '0' when WR_BUF_N = '0' and WR_BUF_SEL = "10" else '1';
RAM_X1_WR_N_2 <= '0' when LayNextWrN = '0' and LayNextWrSel = "10" else '1';

---------------------------------------------------------
RAM_X1_3 : SRAM_64x2560 port map (
   WRCLK	=> CLK,
   RDCLK	=> CLK,
   WR_N		=> RAM_X1_WR_N_3, 
   WRADDR	=> RAM_X1_WR_ADDR, 
   RDADDR	=> RAM_X1_RD_ADDR, 
   WRDATA	=> LayNextWrData, -- RAM_X1_WR_DATA, 
   RDDATA	=> RAM_X1_RD_DATA_3);

-- RAM_X1_WR_N_3 <= '0' when WR_BUF_N = '0' and WR_BUF_SEL = "11" else '1';
RAM_X1_WR_N_3 <= '0' when LayNextWrN = '0' and LayNextWrSel = "11" else '1';

---------------------------------------------------------
RAM_X1_WR_ADDR <= LayNextWrAddr;
RAM_X1_RD_ADDR <= RD_BUF_ADDR(5 downto 0);

RAM_X1_RD_SEL <= RD_BUF_SEL - 2; -- RAM_X1 is written from '3'

RAM_X1_RD_DATA <= RAM_X1_RD_DATA_0 when RAM_X1_RD_SEL = "00" else 
		RAM_X1_RD_DATA_1_padding when RAM_X1_RD_SEL = "01" else 
		RAM_X1_RD_DATA_2_padding when RAM_X1_RD_SEL = "10" else 
		RAM_X1_RD_DATA_3;

RAM_X1_RD_DATA_2_padding <= (others => '0') when SsdLineCount = 3 else RAM_X1_RD_DATA_2; 

RAM_X1_RD_DATA_1_padding <= (others => '0') when SsdLineCount = 40 or SsdLineCount = 41 else RAM_X1_RD_DATA_1;

conv_layer9_inst : conv_layer9 port map (
          CLK => CLK,
          RST_N => RST_N,
          OutChSel => OutChSel,
          CO300 => SsdLineCount,
          CO9 => CO9,
          CO512 => CO512,
          CO512_dly2 => CO512_dly2,
          X_RD_DATA => RAM_X1_RD_DATA,
          ReluAll => ReluAll,
          ReluAllWrN => ReluAllWrN,
          ReluSum => open);

RAM_X2_WR_ADDR <= SsdPixCount(5 downto 0);
--RAM_X2_WR_ADDR <= Pooladdr; -- layer4 has poolings
RAM_X2_WR_DATA <= ReluAll;
--RAM_X2_WR_DATA <= Poolout; -- layer4 has poolings

--tmpout000 <= Poolout(1279 downto 1275);
--tmpout001 <= Poolout(1274 downto 1270);

ReluWrN_tmp <= '0' when ReluAllWrN = '0'  
		and (SsdLineCount2 >= 3 and SsdLineCount2 <= 40) 
		else '1'; 

ReluWrN_sig <= '0' when ReluWrN_tmp = '0' and SsdPixCount >= 1 and SsdPixCount <= 38 else '1';

--process( RST_N, CLK )  
--begin 
--   if( RST_N = '0' ) then
--     ReluWrN_sig <= '0';
--   elsif( CLK'event and CLK = '1' ) then
--     ReluWrN_sig <= ReluWrN_tmp; -- to catch addres '38'(half period)
--  end if;
--end process;

---------------------------------------------------------
RAM_X2_0 : SRAM_64x2560 port map (
   WRCLK	=> CLK,
   RDCLK	=> CLK,
   WR_N		=> RAM_X2_WR_N_0, 
   WRADDR	=> RAM_X2_WR_ADDR, 
   RDADDR	=> RAM_X2_RD_ADDR, 
   WRDATA	=> RAM_X2_WR_DATA, 
   RDDATA	=> RAM_X2_RD_DATA_0);

RAM_X2_WR_N_0 <= '0' when ReluWrN_sig = '0' and WR_BUF_SEL = "00"  
		else '1'; 

---------------------------------------------------------
RAM_X2_1 : SRAM_64x2560 port map (
   WRCLK	=> CLK,
   RDCLK	=> CLK,
   WR_N		=> RAM_X2_WR_N_1, 
   WRADDR	=> RAM_X2_WR_ADDR, 
   RDADDR	=> RAM_X2_RD_ADDR, 
   WRDATA	=> RAM_X2_WR_DATA, 
   RDDATA	=> RAM_X2_RD_DATA_1);

RAM_X2_WR_N_1 <= '0' when ReluWrN_sig = '0' and WR_BUF_SEL = "01" 
		else '1'; 

---------------------------------------------------------
RAM_X2_2 : SRAM_64x2560 port map (
   WRCLK	=> CLK,
   RDCLK	=> CLK,
   WR_N		=> RAM_X2_WR_N_2, 
   WRADDR	=> RAM_X2_WR_ADDR, 
   RDADDR	=> RAM_X2_RD_ADDR, 
   WRDATA	=> RAM_X2_WR_DATA, 
   RDDATA	=> RAM_X2_RD_DATA_2);

RAM_X2_WR_N_2 <= '0' when ReluWrN_sig = '0' and WR_BUF_SEL = "10" 
		else '1'; 

---------------------------------------------------------
RAM_X2_3 : SRAM_64x2560 port map (
   WRCLK	=> CLK,
   RDCLK	=> CLK,
   WR_N		=> RAM_X2_WR_N_3, 
   WRADDR	=> RAM_X2_WR_ADDR, 
   RDADDR	=> RAM_X2_RD_ADDR, 
   WRDATA	=> RAM_X2_WR_DATA, 
   RDDATA	=> RAM_X2_RD_DATA_3);

RAM_X2_WR_N_3 <= '0' when ReluWrN_sig = '0' and WR_BUF_SEL = "11" 
		else '1'; 

---------------------------------------------------------
--pooling_lay7_inst : pooling_lay7 port map (
--   CLK	=> CLK,
--   RST_N	=> RST_N,
--   ReluAll	=> ReluAll,
--   WR_N	=> ReluAllWrN,
--   HCO75	=> HCO75,
--   HCOEDGE	=> HCOEDGE,
--   Poolout	=> Poolout);

HCO75 <= SsdPixCount(6 downto 0) - 0;

Pooladdr <= SsdPixCount2(7 downto 1) + 1; -- 2022/06/24 for pooling SsdPixCount2(7 downto 1) + 1;

---------------------------------------------------------
process( RST_N, CLK )  
begin 
	if( RST_N = '0' ) then
		hco75is75_dly2 <= '0';
		hco75is75_dly1 <= '0';
	elsif( CLK'event and CLK = '1' ) then
		hco75is75_dly2 <= hco75is75_dly1;
		hco75is75_dly1 <= hco75is75;
	end if;  
end process;

hco75is75 <= '1' when HCO75 = 75 else '0';

HCOEDGE <= '1' when hco75is75_dly2 = '1' and hco75is75_dly1 = '0' else '0';

end RTL;

