-- The layer9 of Convolutional Deep Learning on FPGA
-- by Toshio Iwata at digitalfilter.com 2022/01/02

library ieee ;
use ieee.std_logic_1164.all ;
use ieee.std_logic_arith.all ;
use ieee.std_logic_signed.all;

-------------------------------------------------------------------------------
-- Entity
-------------------------------------------------------------------------------
entity conv_layer9 is
port(
  CLK : in std_logic;                            -- clock (100MHz)
  RST_N : in std_logic;                          -- reset (active low)
  OutChSel : in std_logic_vector(1 downto 0);
  CO300 : in std_logic_vector(8 downto 0);
  CO9 : in std_logic_vector(3 downto 0);
  CO512 : in std_logic_vector(8 downto 0);
  CO512_dly2 : in std_logic_vector(8 downto 0);
  X_RD_DATA : in std_logic_vector(2559 downto 0); 
  ReluAll : out std_logic_vector(2559 downto 0); 
  ReluAllWrN : out std_logic; 
  ReluSum : out std_logic_vector(11 downto 0) ); 
end conv_layer9;

-------------------------------------------------------------------------------
-- Architecture
-------------------------------------------------------------------------------
architecture RTL of conv_layer9 is

component dotpro_w9
Port (
  CLK : in std_logic;                            -- clock (100MHz)
  RST_N : in std_logic;                          -- reset (active low)
  PixRomData : in std_logic_vector(4 downto 0); 
  MatRomData : in std_logic_vector(2 downto 0);
  CO9is1 : in std_logic;
  CO512is0 : in std_logic;
  MATMULOUT : Out std_logic_vector(13 downto 0) );  
end component;

component ROM_W9
Port (
          CLK : In std_logic;
          ADDR : In std_logic_vector(12 downto 0);
          DATA : Out std_logic_vector(1535 downto 0) );
end component;

-------------------------------------------------------------------------------
-- Signal
-------------------------------------------------------------------------------
signal CO9_clr : std_logic;
signal W3BUSY, W3BUSY_dly1, W3BUSY_dly2, W3BUSY_dly3, W3BUSY_dly4 : std_logic;

signal mac_000, mac_001, mac_002, mac_003, mac_004, mac_005, mac_006, mac_007, mac_008, mac_009 : std_logic_vector(13 downto 0);
signal mac_010, mac_011, mac_012, mac_013, mac_014, mac_015, mac_016, mac_017, mac_018, mac_019 : std_logic_vector(13 downto 0);
signal mac_020, mac_021, mac_022, mac_023, mac_024, mac_025, mac_026, mac_027, mac_028, mac_029 : std_logic_vector(13 downto 0);
signal mac_030, mac_031, mac_032, mac_033, mac_034, mac_035, mac_036, mac_037, mac_038, mac_039 : std_logic_vector(13 downto 0);
signal mac_040, mac_041, mac_042, mac_043, mac_044, mac_045, mac_046, mac_047, mac_048, mac_049 : std_logic_vector(13 downto 0);
signal mac_050, mac_051, mac_052, mac_053, mac_054, mac_055, mac_056, mac_057, mac_058, mac_059 : std_logic_vector(13 downto 0);
signal mac_060, mac_061, mac_062, mac_063, mac_064, mac_065, mac_066, mac_067, mac_068, mac_069 : std_logic_vector(13 downto 0);
signal mac_070, mac_071, mac_072, mac_073, mac_074, mac_075, mac_076, mac_077, mac_078, mac_079 : std_logic_vector(13 downto 0);
signal mac_080, mac_081, mac_082, mac_083, mac_084, mac_085, mac_086, mac_087, mac_088, mac_089 : std_logic_vector(13 downto 0);
signal mac_090, mac_091, mac_092, mac_093, mac_094, mac_095, mac_096, mac_097, mac_098, mac_099 : std_logic_vector(13 downto 0);
signal mac_100, mac_101, mac_102, mac_103, mac_104, mac_105, mac_106, mac_107, mac_108, mac_109 : std_logic_vector(13 downto 0);
signal mac_110, mac_111, mac_112, mac_113, mac_114, mac_115, mac_116, mac_117, mac_118, mac_119 : std_logic_vector(13 downto 0);
signal mac_120, mac_121, mac_122, mac_123, mac_124, mac_125, mac_126, mac_127 : std_logic_vector(13 downto 0);

signal truncate_000, truncate_001, truncate_002, truncate_003, truncate_004, truncate_005, truncate_006, truncate_007, truncate_008, truncate_009 : std_logic_vector(6 downto 0);
signal truncate_010, truncate_011, truncate_012, truncate_013, truncate_014, truncate_015, truncate_016, truncate_017, truncate_018, truncate_019 : std_logic_vector(6 downto 0);
signal truncate_020, truncate_021, truncate_022, truncate_023, truncate_024, truncate_025, truncate_026, truncate_027, truncate_028, truncate_029 : std_logic_vector(6 downto 0);
signal truncate_030, truncate_031, truncate_032, truncate_033, truncate_034, truncate_035, truncate_036, truncate_037, truncate_038, truncate_039 : std_logic_vector(6 downto 0);
signal truncate_040, truncate_041, truncate_042, truncate_043, truncate_044, truncate_045, truncate_046, truncate_047, truncate_048, truncate_049 : std_logic_vector(6 downto 0);
signal truncate_050, truncate_051, truncate_052, truncate_053, truncate_054, truncate_055, truncate_056, truncate_057, truncate_058, truncate_059 : std_logic_vector(6 downto 0);
signal truncate_060, truncate_061, truncate_062, truncate_063, truncate_064, truncate_065, truncate_066, truncate_067, truncate_068, truncate_069 : std_logic_vector(6 downto 0);
signal truncate_070, truncate_071, truncate_072, truncate_073, truncate_074, truncate_075, truncate_076, truncate_077, truncate_078, truncate_079 : std_logic_vector(6 downto 0);
signal truncate_080, truncate_081, truncate_082, truncate_083, truncate_084, truncate_085, truncate_086, truncate_087, truncate_088, truncate_089 : std_logic_vector(6 downto 0);
signal truncate_090, truncate_091, truncate_092, truncate_093, truncate_094, truncate_095, truncate_096, truncate_097, truncate_098, truncate_099 : std_logic_vector(6 downto 0);
signal truncate_100, truncate_101, truncate_102, truncate_103, truncate_104, truncate_105, truncate_106, truncate_107, truncate_108, truncate_109 : std_logic_vector(6 downto 0);
signal truncate_110, truncate_111, truncate_112, truncate_113, truncate_114, truncate_115, truncate_116, truncate_117, truncate_118, truncate_119 : std_logic_vector(6 downto 0);
signal truncate_120, truncate_121, truncate_122, truncate_123, truncate_124, truncate_125, truncate_126, truncate_127 : std_logic_vector(6 downto 0);

signal Relu_000, Relu_001, Relu_002, Relu_003, Relu_004, Relu_005, Relu_006, Relu_007, Relu_008, Relu_009 : std_logic_vector(4 downto 0);
signal Relu_010, Relu_011, Relu_012, Relu_013, Relu_014, Relu_015, Relu_016, Relu_017, Relu_018, Relu_019 : std_logic_vector(4 downto 0);
signal Relu_020, Relu_021, Relu_022, Relu_023, Relu_024, Relu_025, Relu_026, Relu_027, Relu_028, Relu_029 : std_logic_vector(4 downto 0);
signal Relu_030, Relu_031, Relu_032, Relu_033, Relu_034, Relu_035, Relu_036, Relu_037, Relu_038, Relu_039 : std_logic_vector(4 downto 0);
signal Relu_040, Relu_041, Relu_042, Relu_043, Relu_044, Relu_045, Relu_046, Relu_047, Relu_048, Relu_049 : std_logic_vector(4 downto 0);
signal Relu_050, Relu_051, Relu_052, Relu_053, Relu_054, Relu_055, Relu_056, Relu_057, Relu_058, Relu_059 : std_logic_vector(4 downto 0);
signal Relu_060, Relu_061, Relu_062, Relu_063, Relu_064, Relu_065, Relu_066, Relu_067, Relu_068, Relu_069 : std_logic_vector(4 downto 0);
signal Relu_070, Relu_071, Relu_072, Relu_073, Relu_074, Relu_075, Relu_076, Relu_077, Relu_078, Relu_079 : std_logic_vector(4 downto 0);
signal Relu_080, Relu_081, Relu_082, Relu_083, Relu_084, Relu_085, Relu_086, Relu_087, Relu_088, Relu_089 : std_logic_vector(4 downto 0);
signal Relu_090, Relu_091, Relu_092, Relu_093, Relu_094, Relu_095, Relu_096, Relu_097, Relu_098, Relu_099 : std_logic_vector(4 downto 0);
signal Relu_100, Relu_101, Relu_102, Relu_103, Relu_104, Relu_105, Relu_106, Relu_107, Relu_108, Relu_109 : std_logic_vector(4 downto 0);
signal Relu_110, Relu_111, Relu_112, Relu_113, Relu_114, Relu_115, Relu_116, Relu_117, Relu_118, Relu_119 : std_logic_vector(4 downto 0);
signal Relu_120, Relu_121, Relu_122, Relu_123, Relu_124, Relu_125, Relu_126, Relu_127 : std_logic_vector(4 downto 0);

signal PixRomData : std_logic_vector(4 downto 0);
signal W9Addr, W9Addr_prev : std_logic_vector(12 downto 0);
signal W9Addr_L : std_logic_vector(3 downto 0);
signal W9Addr_H : std_logic_vector(12 downto 0);
signal W9Sel : std_logic_vector(1 downto 0);
signal OutChSel_dly1, OutChSel_dly2 : std_logic_vector(1 downto 0);
signal reg_000_0, reg_000_1, reg_000_2, reg_000_3: std_logic_vector(4 downto 0);
signal reg_001_0, reg_001_1, reg_001_2, reg_001_3: std_logic_vector(4 downto 0);
signal reg_002_0, reg_002_1, reg_002_2, reg_002_3: std_logic_vector(4 downto 0);
signal reg_003_0, reg_003_1, reg_003_2, reg_003_3: std_logic_vector(4 downto 0);
signal reg_004_0, reg_004_1, reg_004_2, reg_004_3: std_logic_vector(4 downto 0);
signal reg_005_0, reg_005_1, reg_005_2, reg_005_3: std_logic_vector(4 downto 0);
signal reg_006_0, reg_006_1, reg_006_2, reg_006_3: std_logic_vector(4 downto 0);
signal reg_007_0, reg_007_1, reg_007_2, reg_007_3: std_logic_vector(4 downto 0);
signal reg_008_0, reg_008_1, reg_008_2, reg_008_3: std_logic_vector(4 downto 0);
signal reg_009_0, reg_009_1, reg_009_2, reg_009_3: std_logic_vector(4 downto 0);
signal reg_010_0, reg_010_1, reg_010_2, reg_010_3: std_logic_vector(4 downto 0);
signal reg_011_0, reg_011_1, reg_011_2, reg_011_3: std_logic_vector(4 downto 0);
signal reg_012_0, reg_012_1, reg_012_2, reg_012_3: std_logic_vector(4 downto 0);
signal reg_013_0, reg_013_1, reg_013_2, reg_013_3: std_logic_vector(4 downto 0);
signal reg_014_0, reg_014_1, reg_014_2, reg_014_3: std_logic_vector(4 downto 0);
signal reg_015_0, reg_015_1, reg_015_2, reg_015_3: std_logic_vector(4 downto 0);
signal reg_016_0, reg_016_1, reg_016_2, reg_016_3: std_logic_vector(4 downto 0);
signal reg_017_0, reg_017_1, reg_017_2, reg_017_3: std_logic_vector(4 downto 0);
signal reg_018_0, reg_018_1, reg_018_2, reg_018_3: std_logic_vector(4 downto 0);
signal reg_019_0, reg_019_1, reg_019_2, reg_019_3: std_logic_vector(4 downto 0);
signal reg_020_0, reg_020_1, reg_020_2, reg_020_3: std_logic_vector(4 downto 0);
signal reg_021_0, reg_021_1, reg_021_2, reg_021_3: std_logic_vector(4 downto 0);
signal reg_022_0, reg_022_1, reg_022_2, reg_022_3: std_logic_vector(4 downto 0);
signal reg_023_0, reg_023_1, reg_023_2, reg_023_3: std_logic_vector(4 downto 0);
signal reg_024_0, reg_024_1, reg_024_2, reg_024_3: std_logic_vector(4 downto 0);
signal reg_025_0, reg_025_1, reg_025_2, reg_025_3: std_logic_vector(4 downto 0);
signal reg_026_0, reg_026_1, reg_026_2, reg_026_3: std_logic_vector(4 downto 0);
signal reg_027_0, reg_027_1, reg_027_2, reg_027_3: std_logic_vector(4 downto 0);
signal reg_028_0, reg_028_1, reg_028_2, reg_028_3: std_logic_vector(4 downto 0);
signal reg_029_0, reg_029_1, reg_029_2, reg_029_3: std_logic_vector(4 downto 0);
signal reg_030_0, reg_030_1, reg_030_2, reg_030_3: std_logic_vector(4 downto 0);
signal reg_031_0, reg_031_1, reg_031_2, reg_031_3: std_logic_vector(4 downto 0);
signal reg_032_0, reg_032_1, reg_032_2, reg_032_3: std_logic_vector(4 downto 0);
signal reg_033_0, reg_033_1, reg_033_2, reg_033_3: std_logic_vector(4 downto 0);
signal reg_034_0, reg_034_1, reg_034_2, reg_034_3: std_logic_vector(4 downto 0);
signal reg_035_0, reg_035_1, reg_035_2, reg_035_3: std_logic_vector(4 downto 0);
signal reg_036_0, reg_036_1, reg_036_2, reg_036_3: std_logic_vector(4 downto 0);
signal reg_037_0, reg_037_1, reg_037_2, reg_037_3: std_logic_vector(4 downto 0);
signal reg_038_0, reg_038_1, reg_038_2, reg_038_3: std_logic_vector(4 downto 0);
signal reg_039_0, reg_039_1, reg_039_2, reg_039_3: std_logic_vector(4 downto 0);
signal reg_040_0, reg_040_1, reg_040_2, reg_040_3: std_logic_vector(4 downto 0);
signal reg_041_0, reg_041_1, reg_041_2, reg_041_3: std_logic_vector(4 downto 0);
signal reg_042_0, reg_042_1, reg_042_2, reg_042_3: std_logic_vector(4 downto 0);
signal reg_043_0, reg_043_1, reg_043_2, reg_043_3: std_logic_vector(4 downto 0);
signal reg_044_0, reg_044_1, reg_044_2, reg_044_3: std_logic_vector(4 downto 0);
signal reg_045_0, reg_045_1, reg_045_2, reg_045_3: std_logic_vector(4 downto 0);
signal reg_046_0, reg_046_1, reg_046_2, reg_046_3: std_logic_vector(4 downto 0);
signal reg_047_0, reg_047_1, reg_047_2, reg_047_3: std_logic_vector(4 downto 0);
signal reg_048_0, reg_048_1, reg_048_2, reg_048_3: std_logic_vector(4 downto 0);
signal reg_049_0, reg_049_1, reg_049_2, reg_049_3: std_logic_vector(4 downto 0);
signal reg_050_0, reg_050_1, reg_050_2, reg_050_3: std_logic_vector(4 downto 0);
signal reg_051_0, reg_051_1, reg_051_2, reg_051_3: std_logic_vector(4 downto 0);
signal reg_052_0, reg_052_1, reg_052_2, reg_052_3: std_logic_vector(4 downto 0);
signal reg_053_0, reg_053_1, reg_053_2, reg_053_3: std_logic_vector(4 downto 0);
signal reg_054_0, reg_054_1, reg_054_2, reg_054_3: std_logic_vector(4 downto 0);
signal reg_055_0, reg_055_1, reg_055_2, reg_055_3: std_logic_vector(4 downto 0);
signal reg_056_0, reg_056_1, reg_056_2, reg_056_3: std_logic_vector(4 downto 0);
signal reg_057_0, reg_057_1, reg_057_2, reg_057_3: std_logic_vector(4 downto 0);
signal reg_058_0, reg_058_1, reg_058_2, reg_058_3: std_logic_vector(4 downto 0);
signal reg_059_0, reg_059_1, reg_059_2, reg_059_3: std_logic_vector(4 downto 0);
signal reg_060_0, reg_060_1, reg_060_2, reg_060_3: std_logic_vector(4 downto 0);
signal reg_061_0, reg_061_1, reg_061_2, reg_061_3: std_logic_vector(4 downto 0);
signal reg_062_0, reg_062_1, reg_062_2, reg_062_3: std_logic_vector(4 downto 0);
signal reg_063_0, reg_063_1, reg_063_2, reg_063_3: std_logic_vector(4 downto 0);
signal reg_064_0, reg_064_1, reg_064_2, reg_064_3: std_logic_vector(4 downto 0);
signal reg_065_0, reg_065_1, reg_065_2, reg_065_3: std_logic_vector(4 downto 0);
signal reg_066_0, reg_066_1, reg_066_2, reg_066_3: std_logic_vector(4 downto 0);
signal reg_067_0, reg_067_1, reg_067_2, reg_067_3: std_logic_vector(4 downto 0);
signal reg_068_0, reg_068_1, reg_068_2, reg_068_3: std_logic_vector(4 downto 0);
signal reg_069_0, reg_069_1, reg_069_2, reg_069_3: std_logic_vector(4 downto 0);
signal reg_070_0, reg_070_1, reg_070_2, reg_070_3: std_logic_vector(4 downto 0);
signal reg_071_0, reg_071_1, reg_071_2, reg_071_3: std_logic_vector(4 downto 0);
signal reg_072_0, reg_072_1, reg_072_2, reg_072_3: std_logic_vector(4 downto 0);
signal reg_073_0, reg_073_1, reg_073_2, reg_073_3: std_logic_vector(4 downto 0);
signal reg_074_0, reg_074_1, reg_074_2, reg_074_3: std_logic_vector(4 downto 0);
signal reg_075_0, reg_075_1, reg_075_2, reg_075_3: std_logic_vector(4 downto 0);
signal reg_076_0, reg_076_1, reg_076_2, reg_076_3: std_logic_vector(4 downto 0);
signal reg_077_0, reg_077_1, reg_077_2, reg_077_3: std_logic_vector(4 downto 0);
signal reg_078_0, reg_078_1, reg_078_2, reg_078_3: std_logic_vector(4 downto 0);
signal reg_079_0, reg_079_1, reg_079_2, reg_079_3: std_logic_vector(4 downto 0);
signal reg_080_0, reg_080_1, reg_080_2, reg_080_3: std_logic_vector(4 downto 0);
signal reg_081_0, reg_081_1, reg_081_2, reg_081_3: std_logic_vector(4 downto 0);
signal reg_082_0, reg_082_1, reg_082_2, reg_082_3: std_logic_vector(4 downto 0);
signal reg_083_0, reg_083_1, reg_083_2, reg_083_3: std_logic_vector(4 downto 0);
signal reg_084_0, reg_084_1, reg_084_2, reg_084_3: std_logic_vector(4 downto 0);
signal reg_085_0, reg_085_1, reg_085_2, reg_085_3: std_logic_vector(4 downto 0);
signal reg_086_0, reg_086_1, reg_086_2, reg_086_3: std_logic_vector(4 downto 0);
signal reg_087_0, reg_087_1, reg_087_2, reg_087_3: std_logic_vector(4 downto 0);
signal reg_088_0, reg_088_1, reg_088_2, reg_088_3: std_logic_vector(4 downto 0);
signal reg_089_0, reg_089_1, reg_089_2, reg_089_3: std_logic_vector(4 downto 0);
signal reg_090_0, reg_090_1, reg_090_2, reg_090_3: std_logic_vector(4 downto 0);
signal reg_091_0, reg_091_1, reg_091_2, reg_091_3: std_logic_vector(4 downto 0);
signal reg_092_0, reg_092_1, reg_092_2, reg_092_3: std_logic_vector(4 downto 0);
signal reg_093_0, reg_093_1, reg_093_2, reg_093_3: std_logic_vector(4 downto 0);
signal reg_094_0, reg_094_1, reg_094_2, reg_094_3: std_logic_vector(4 downto 0);
signal reg_095_0, reg_095_1, reg_095_2, reg_095_3: std_logic_vector(4 downto 0);
signal reg_096_0, reg_096_1, reg_096_2, reg_096_3: std_logic_vector(4 downto 0);
signal reg_097_0, reg_097_1, reg_097_2, reg_097_3: std_logic_vector(4 downto 0);
signal reg_098_0, reg_098_1, reg_098_2, reg_098_3: std_logic_vector(4 downto 0);
signal reg_099_0, reg_099_1, reg_099_2, reg_099_3: std_logic_vector(4 downto 0);
signal reg_100_0, reg_100_1, reg_100_2, reg_100_3: std_logic_vector(4 downto 0);
signal reg_101_0, reg_101_1, reg_101_2, reg_101_3: std_logic_vector(4 downto 0);
signal reg_102_0, reg_102_1, reg_102_2, reg_102_3: std_logic_vector(4 downto 0);
signal reg_103_0, reg_103_1, reg_103_2, reg_103_3: std_logic_vector(4 downto 0);
signal reg_104_0, reg_104_1, reg_104_2, reg_104_3: std_logic_vector(4 downto 0);
signal reg_105_0, reg_105_1, reg_105_2, reg_105_3: std_logic_vector(4 downto 0);
signal reg_106_0, reg_106_1, reg_106_2, reg_106_3: std_logic_vector(4 downto 0);
signal reg_107_0, reg_107_1, reg_107_2, reg_107_3: std_logic_vector(4 downto 0);
signal reg_108_0, reg_108_1, reg_108_2, reg_108_3: std_logic_vector(4 downto 0);
signal reg_109_0, reg_109_1, reg_109_2, reg_109_3: std_logic_vector(4 downto 0);
signal reg_110_0, reg_110_1, reg_110_2, reg_110_3: std_logic_vector(4 downto 0);
signal reg_111_0, reg_111_1, reg_111_2, reg_111_3: std_logic_vector(4 downto 0);
signal reg_112_0, reg_112_1, reg_112_2, reg_112_3: std_logic_vector(4 downto 0);
signal reg_113_0, reg_113_1, reg_113_2, reg_113_3: std_logic_vector(4 downto 0);
signal reg_114_0, reg_114_1, reg_114_2, reg_114_3: std_logic_vector(4 downto 0);
signal reg_115_0, reg_115_1, reg_115_2, reg_115_3: std_logic_vector(4 downto 0);
signal reg_116_0, reg_116_1, reg_116_2, reg_116_3: std_logic_vector(4 downto 0);
signal reg_117_0, reg_117_1, reg_117_2, reg_117_3: std_logic_vector(4 downto 0);
signal reg_118_0, reg_118_1, reg_118_2, reg_118_3: std_logic_vector(4 downto 0);
signal reg_119_0, reg_119_1, reg_119_2, reg_119_3: std_logic_vector(4 downto 0);
signal reg_120_0, reg_120_1, reg_120_2, reg_120_3: std_logic_vector(4 downto 0);
signal reg_121_0, reg_121_1, reg_121_2, reg_121_3: std_logic_vector(4 downto 0);
signal reg_122_0, reg_122_1, reg_122_2, reg_122_3: std_logic_vector(4 downto 0);
signal reg_123_0, reg_123_1, reg_123_2, reg_123_3: std_logic_vector(4 downto 0);
signal reg_124_0, reg_124_1, reg_124_2, reg_124_3: std_logic_vector(4 downto 0);
signal reg_125_0, reg_125_1, reg_125_2, reg_125_3: std_logic_vector(4 downto 0);
signal reg_126_0, reg_126_1, reg_126_2, reg_126_3: std_logic_vector(4 downto 0);
signal reg_127_0, reg_127_1, reg_127_2, reg_127_3: std_logic_vector(4 downto 0);
signal enb_a, enb_b, enb_c, enb_d : std_logic;
signal CO9is1, CO512is0 : std_logic;
signal W9Data : std_logic_vector(1535 downto 0);
signal coeff_000, coeff_001, coeff_002, coeff_003, coeff_004, coeff_005, coeff_006, coeff_007, coeff_008, coeff_009: std_logic_vector(2 downto 0);
signal coeff_010, coeff_011, coeff_012, coeff_013, coeff_014, coeff_015, coeff_016, coeff_017, coeff_018, coeff_019: std_logic_vector(2 downto 0);
signal coeff_020, coeff_021, coeff_022, coeff_023, coeff_024, coeff_025, coeff_026, coeff_027, coeff_028, coeff_029: std_logic_vector(2 downto 0);
signal coeff_030, coeff_031, coeff_032, coeff_033, coeff_034, coeff_035, coeff_036, coeff_037, coeff_038, coeff_039: std_logic_vector(2 downto 0);
signal coeff_040, coeff_041, coeff_042, coeff_043, coeff_044, coeff_045, coeff_046, coeff_047, coeff_048, coeff_049: std_logic_vector(2 downto 0);
signal coeff_050, coeff_051, coeff_052, coeff_053, coeff_054, coeff_055, coeff_056, coeff_057, coeff_058, coeff_059: std_logic_vector(2 downto 0);
signal coeff_060, coeff_061, coeff_062, coeff_063, coeff_064, coeff_065, coeff_066, coeff_067, coeff_068, coeff_069: std_logic_vector(2 downto 0);
signal coeff_070, coeff_071, coeff_072, coeff_073, coeff_074, coeff_075, coeff_076, coeff_077, coeff_078, coeff_079: std_logic_vector(2 downto 0);
signal coeff_080, coeff_081, coeff_082, coeff_083, coeff_084, coeff_085, coeff_086, coeff_087, coeff_088, coeff_089: std_logic_vector(2 downto 0);
signal coeff_090, coeff_091, coeff_092, coeff_093, coeff_094, coeff_095, coeff_096, coeff_097, coeff_098, coeff_099: std_logic_vector(2 downto 0);
signal coeff_100, coeff_101, coeff_102, coeff_103, coeff_104, coeff_105, coeff_106, coeff_107, coeff_108, coeff_109: std_logic_vector(2 downto 0);
signal coeff_110, coeff_111, coeff_112, coeff_113, coeff_114, coeff_115, coeff_116, coeff_117, coeff_118, coeff_119: std_logic_vector(2 downto 0);
signal coeff_120, coeff_121, coeff_122, coeff_123, coeff_124, coeff_125, coeff_126, coeff_127: std_logic_vector(2 downto 0);

begin

CO9is1 <= '1' when CO9 = 1 else '0'; 
CO512is0 <= '1' when CO512 = 0 else '0';

---------------------------------------------------------
process( RST_N, CLK )  
begin 
	if( RST_N = '0' ) then
		OutChSel_dly2 <= (others => '0');
		OutChSel_dly1 <= (others => '0');
	elsif( CLK'event and CLK = '1' ) then
		OutChSel_dly2 <= OutChSel_dly1;
		OutChSel_dly1 <= OutChSel;
	end if;  
end process;

W9Sel <= OutChSel_dly2(1 downto 0);

---------------------------------------------------------
  PixRomData <= 
X_RD_DATA( 2559 downto 2555 ) when CO512_dly2 = "000000000" else 
X_RD_DATA( 2554 downto 2550 ) when CO512_dly2 = "000000001" else 
X_RD_DATA( 2549 downto 2545 ) when CO512_dly2 = "000000010" else 
X_RD_DATA( 2544 downto 2540 ) when CO512_dly2 = "000000011" else 
X_RD_DATA( 2539 downto 2535 ) when CO512_dly2 = "000000100" else 
X_RD_DATA( 2534 downto 2530 ) when CO512_dly2 = "000000101" else 
X_RD_DATA( 2529 downto 2525 ) when CO512_dly2 = "000000110" else 
X_RD_DATA( 2524 downto 2520 ) when CO512_dly2 = "000000111" else 
X_RD_DATA( 2519 downto 2515 ) when CO512_dly2 = "000001000" else 
X_RD_DATA( 2514 downto 2510 ) when CO512_dly2 = "000001001" else 
X_RD_DATA( 2509 downto 2505 ) when CO512_dly2 = "000001010" else 
X_RD_DATA( 2504 downto 2500 ) when CO512_dly2 = "000001011" else 
X_RD_DATA( 2499 downto 2495 ) when CO512_dly2 = "000001100" else 
X_RD_DATA( 2494 downto 2490 ) when CO512_dly2 = "000001101" else 
X_RD_DATA( 2489 downto 2485 ) when CO512_dly2 = "000001110" else 
X_RD_DATA( 2484 downto 2480 ) when CO512_dly2 = "000001111" else 
X_RD_DATA( 2479 downto 2475 ) when CO512_dly2 = "000010000" else 
X_RD_DATA( 2474 downto 2470 ) when CO512_dly2 = "000010001" else 
X_RD_DATA( 2469 downto 2465 ) when CO512_dly2 = "000010010" else 
X_RD_DATA( 2464 downto 2460 ) when CO512_dly2 = "000010011" else 
X_RD_DATA( 2459 downto 2455 ) when CO512_dly2 = "000010100" else 
X_RD_DATA( 2454 downto 2450 ) when CO512_dly2 = "000010101" else 
X_RD_DATA( 2449 downto 2445 ) when CO512_dly2 = "000010110" else 
X_RD_DATA( 2444 downto 2440 ) when CO512_dly2 = "000010111" else 
X_RD_DATA( 2439 downto 2435 ) when CO512_dly2 = "000011000" else 
X_RD_DATA( 2434 downto 2430 ) when CO512_dly2 = "000011001" else 
X_RD_DATA( 2429 downto 2425 ) when CO512_dly2 = "000011010" else 
X_RD_DATA( 2424 downto 2420 ) when CO512_dly2 = "000011011" else 
X_RD_DATA( 2419 downto 2415 ) when CO512_dly2 = "000011100" else 
X_RD_DATA( 2414 downto 2410 ) when CO512_dly2 = "000011101" else 
X_RD_DATA( 2409 downto 2405 ) when CO512_dly2 = "000011110" else 
X_RD_DATA( 2404 downto 2400 ) when CO512_dly2 = "000011111" else 
X_RD_DATA( 2399 downto 2395 ) when CO512_dly2 = "000100000" else 
X_RD_DATA( 2394 downto 2390 ) when CO512_dly2 = "000100001" else 
X_RD_DATA( 2389 downto 2385 ) when CO512_dly2 = "000100010" else 
X_RD_DATA( 2384 downto 2380 ) when CO512_dly2 = "000100011" else 
X_RD_DATA( 2379 downto 2375 ) when CO512_dly2 = "000100100" else 
X_RD_DATA( 2374 downto 2370 ) when CO512_dly2 = "000100101" else 
X_RD_DATA( 2369 downto 2365 ) when CO512_dly2 = "000100110" else 
X_RD_DATA( 2364 downto 2360 ) when CO512_dly2 = "000100111" else 
X_RD_DATA( 2359 downto 2355 ) when CO512_dly2 = "000101000" else 
X_RD_DATA( 2354 downto 2350 ) when CO512_dly2 = "000101001" else 
X_RD_DATA( 2349 downto 2345 ) when CO512_dly2 = "000101010" else 
X_RD_DATA( 2344 downto 2340 ) when CO512_dly2 = "000101011" else 
X_RD_DATA( 2339 downto 2335 ) when CO512_dly2 = "000101100" else 
X_RD_DATA( 2334 downto 2330 ) when CO512_dly2 = "000101101" else 
X_RD_DATA( 2329 downto 2325 ) when CO512_dly2 = "000101110" else 
X_RD_DATA( 2324 downto 2320 ) when CO512_dly2 = "000101111" else 
X_RD_DATA( 2319 downto 2315 ) when CO512_dly2 = "000110000" else 
X_RD_DATA( 2314 downto 2310 ) when CO512_dly2 = "000110001" else 
X_RD_DATA( 2309 downto 2305 ) when CO512_dly2 = "000110010" else 
X_RD_DATA( 2304 downto 2300 ) when CO512_dly2 = "000110011" else 
X_RD_DATA( 2299 downto 2295 ) when CO512_dly2 = "000110100" else 
X_RD_DATA( 2294 downto 2290 ) when CO512_dly2 = "000110101" else 
X_RD_DATA( 2289 downto 2285 ) when CO512_dly2 = "000110110" else 
X_RD_DATA( 2284 downto 2280 ) when CO512_dly2 = "000110111" else 
X_RD_DATA( 2279 downto 2275 ) when CO512_dly2 = "000111000" else 
X_RD_DATA( 2274 downto 2270 ) when CO512_dly2 = "000111001" else 
X_RD_DATA( 2269 downto 2265 ) when CO512_dly2 = "000111010" else 
X_RD_DATA( 2264 downto 2260 ) when CO512_dly2 = "000111011" else 
X_RD_DATA( 2259 downto 2255 ) when CO512_dly2 = "000111100" else 
X_RD_DATA( 2254 downto 2250 ) when CO512_dly2 = "000111101" else 
X_RD_DATA( 2249 downto 2245 ) when CO512_dly2 = "000111110" else 
X_RD_DATA( 2244 downto 2240 ) when CO512_dly2 = "000111111" else 
X_RD_DATA( 2239 downto 2235 ) when CO512_dly2 = "001000000" else 
X_RD_DATA( 2234 downto 2230 ) when CO512_dly2 = "001000001" else 
X_RD_DATA( 2229 downto 2225 ) when CO512_dly2 = "001000010" else 
X_RD_DATA( 2224 downto 2220 ) when CO512_dly2 = "001000011" else 
X_RD_DATA( 2219 downto 2215 ) when CO512_dly2 = "001000100" else 
X_RD_DATA( 2214 downto 2210 ) when CO512_dly2 = "001000101" else 
X_RD_DATA( 2209 downto 2205 ) when CO512_dly2 = "001000110" else 
X_RD_DATA( 2204 downto 2200 ) when CO512_dly2 = "001000111" else 
X_RD_DATA( 2199 downto 2195 ) when CO512_dly2 = "001001000" else 
X_RD_DATA( 2194 downto 2190 ) when CO512_dly2 = "001001001" else 
X_RD_DATA( 2189 downto 2185 ) when CO512_dly2 = "001001010" else 
X_RD_DATA( 2184 downto 2180 ) when CO512_dly2 = "001001011" else 
X_RD_DATA( 2179 downto 2175 ) when CO512_dly2 = "001001100" else 
X_RD_DATA( 2174 downto 2170 ) when CO512_dly2 = "001001101" else 
X_RD_DATA( 2169 downto 2165 ) when CO512_dly2 = "001001110" else 
X_RD_DATA( 2164 downto 2160 ) when CO512_dly2 = "001001111" else 
X_RD_DATA( 2159 downto 2155 ) when CO512_dly2 = "001010000" else 
X_RD_DATA( 2154 downto 2150 ) when CO512_dly2 = "001010001" else 
X_RD_DATA( 2149 downto 2145 ) when CO512_dly2 = "001010010" else 
X_RD_DATA( 2144 downto 2140 ) when CO512_dly2 = "001010011" else 
X_RD_DATA( 2139 downto 2135 ) when CO512_dly2 = "001010100" else 
X_RD_DATA( 2134 downto 2130 ) when CO512_dly2 = "001010101" else 
X_RD_DATA( 2129 downto 2125 ) when CO512_dly2 = "001010110" else 
X_RD_DATA( 2124 downto 2120 ) when CO512_dly2 = "001010111" else 
X_RD_DATA( 2119 downto 2115 ) when CO512_dly2 = "001011000" else 
X_RD_DATA( 2114 downto 2110 ) when CO512_dly2 = "001011001" else 
X_RD_DATA( 2109 downto 2105 ) when CO512_dly2 = "001011010" else 
X_RD_DATA( 2104 downto 2100 ) when CO512_dly2 = "001011011" else 
X_RD_DATA( 2099 downto 2095 ) when CO512_dly2 = "001011100" else 
X_RD_DATA( 2094 downto 2090 ) when CO512_dly2 = "001011101" else 
X_RD_DATA( 2089 downto 2085 ) when CO512_dly2 = "001011110" else 
X_RD_DATA( 2084 downto 2080 ) when CO512_dly2 = "001011111" else 
X_RD_DATA( 2079 downto 2075 ) when CO512_dly2 = "001100000" else 
X_RD_DATA( 2074 downto 2070 ) when CO512_dly2 = "001100001" else 
X_RD_DATA( 2069 downto 2065 ) when CO512_dly2 = "001100010" else 
X_RD_DATA( 2064 downto 2060 ) when CO512_dly2 = "001100011" else 
X_RD_DATA( 2059 downto 2055 ) when CO512_dly2 = "001100100" else 
X_RD_DATA( 2054 downto 2050 ) when CO512_dly2 = "001100101" else 
X_RD_DATA( 2049 downto 2045 ) when CO512_dly2 = "001100110" else 
X_RD_DATA( 2044 downto 2040 ) when CO512_dly2 = "001100111" else 
X_RD_DATA( 2039 downto 2035 ) when CO512_dly2 = "001101000" else 
X_RD_DATA( 2034 downto 2030 ) when CO512_dly2 = "001101001" else 
X_RD_DATA( 2029 downto 2025 ) when CO512_dly2 = "001101010" else 
X_RD_DATA( 2024 downto 2020 ) when CO512_dly2 = "001101011" else 
X_RD_DATA( 2019 downto 2015 ) when CO512_dly2 = "001101100" else 
X_RD_DATA( 2014 downto 2010 ) when CO512_dly2 = "001101101" else 
X_RD_DATA( 2009 downto 2005 ) when CO512_dly2 = "001101110" else 
X_RD_DATA( 2004 downto 2000 ) when CO512_dly2 = "001101111" else 
X_RD_DATA( 1999 downto 1995 ) when CO512_dly2 = "001110000" else 
X_RD_DATA( 1994 downto 1990 ) when CO512_dly2 = "001110001" else 
X_RD_DATA( 1989 downto 1985 ) when CO512_dly2 = "001110010" else 
X_RD_DATA( 1984 downto 1980 ) when CO512_dly2 = "001110011" else 
X_RD_DATA( 1979 downto 1975 ) when CO512_dly2 = "001110100" else 
X_RD_DATA( 1974 downto 1970 ) when CO512_dly2 = "001110101" else 
X_RD_DATA( 1969 downto 1965 ) when CO512_dly2 = "001110110" else 
X_RD_DATA( 1964 downto 1960 ) when CO512_dly2 = "001110111" else 
X_RD_DATA( 1959 downto 1955 ) when CO512_dly2 = "001111000" else 
X_RD_DATA( 1954 downto 1950 ) when CO512_dly2 = "001111001" else 
X_RD_DATA( 1949 downto 1945 ) when CO512_dly2 = "001111010" else 
X_RD_DATA( 1944 downto 1940 ) when CO512_dly2 = "001111011" else 
X_RD_DATA( 1939 downto 1935 ) when CO512_dly2 = "001111100" else 
X_RD_DATA( 1934 downto 1930 ) when CO512_dly2 = "001111101" else 
X_RD_DATA( 1929 downto 1925 ) when CO512_dly2 = "001111110" else 
X_RD_DATA( 1924 downto 1920 ) when CO512_dly2 = "001111111" else 
X_RD_DATA( 1919 downto 1915 ) when CO512_dly2 = "010000000" else 
X_RD_DATA( 1914 downto 1910 ) when CO512_dly2 = "010000001" else 
X_RD_DATA( 1909 downto 1905 ) when CO512_dly2 = "010000010" else 
X_RD_DATA( 1904 downto 1900 ) when CO512_dly2 = "010000011" else 
X_RD_DATA( 1899 downto 1895 ) when CO512_dly2 = "010000100" else 
X_RD_DATA( 1894 downto 1890 ) when CO512_dly2 = "010000101" else 
X_RD_DATA( 1889 downto 1885 ) when CO512_dly2 = "010000110" else 
X_RD_DATA( 1884 downto 1880 ) when CO512_dly2 = "010000111" else 
X_RD_DATA( 1879 downto 1875 ) when CO512_dly2 = "010001000" else 
X_RD_DATA( 1874 downto 1870 ) when CO512_dly2 = "010001001" else 
X_RD_DATA( 1869 downto 1865 ) when CO512_dly2 = "010001010" else 
X_RD_DATA( 1864 downto 1860 ) when CO512_dly2 = "010001011" else 
X_RD_DATA( 1859 downto 1855 ) when CO512_dly2 = "010001100" else 
X_RD_DATA( 1854 downto 1850 ) when CO512_dly2 = "010001101" else 
X_RD_DATA( 1849 downto 1845 ) when CO512_dly2 = "010001110" else 
X_RD_DATA( 1844 downto 1840 ) when CO512_dly2 = "010001111" else 
X_RD_DATA( 1839 downto 1835 ) when CO512_dly2 = "010010000" else 
X_RD_DATA( 1834 downto 1830 ) when CO512_dly2 = "010010001" else 
X_RD_DATA( 1829 downto 1825 ) when CO512_dly2 = "010010010" else 
X_RD_DATA( 1824 downto 1820 ) when CO512_dly2 = "010010011" else 
X_RD_DATA( 1819 downto 1815 ) when CO512_dly2 = "010010100" else 
X_RD_DATA( 1814 downto 1810 ) when CO512_dly2 = "010010101" else 
X_RD_DATA( 1809 downto 1805 ) when CO512_dly2 = "010010110" else 
X_RD_DATA( 1804 downto 1800 ) when CO512_dly2 = "010010111" else 
X_RD_DATA( 1799 downto 1795 ) when CO512_dly2 = "010011000" else 
X_RD_DATA( 1794 downto 1790 ) when CO512_dly2 = "010011001" else 
X_RD_DATA( 1789 downto 1785 ) when CO512_dly2 = "010011010" else 
X_RD_DATA( 1784 downto 1780 ) when CO512_dly2 = "010011011" else 
X_RD_DATA( 1779 downto 1775 ) when CO512_dly2 = "010011100" else 
X_RD_DATA( 1774 downto 1770 ) when CO512_dly2 = "010011101" else 
X_RD_DATA( 1769 downto 1765 ) when CO512_dly2 = "010011110" else 
X_RD_DATA( 1764 downto 1760 ) when CO512_dly2 = "010011111" else 
X_RD_DATA( 1759 downto 1755 ) when CO512_dly2 = "010100000" else 
X_RD_DATA( 1754 downto 1750 ) when CO512_dly2 = "010100001" else 
X_RD_DATA( 1749 downto 1745 ) when CO512_dly2 = "010100010" else 
X_RD_DATA( 1744 downto 1740 ) when CO512_dly2 = "010100011" else 
X_RD_DATA( 1739 downto 1735 ) when CO512_dly2 = "010100100" else 
X_RD_DATA( 1734 downto 1730 ) when CO512_dly2 = "010100101" else 
X_RD_DATA( 1729 downto 1725 ) when CO512_dly2 = "010100110" else 
X_RD_DATA( 1724 downto 1720 ) when CO512_dly2 = "010100111" else 
X_RD_DATA( 1719 downto 1715 ) when CO512_dly2 = "010101000" else 
X_RD_DATA( 1714 downto 1710 ) when CO512_dly2 = "010101001" else 
X_RD_DATA( 1709 downto 1705 ) when CO512_dly2 = "010101010" else 
X_RD_DATA( 1704 downto 1700 ) when CO512_dly2 = "010101011" else 
X_RD_DATA( 1699 downto 1695 ) when CO512_dly2 = "010101100" else 
X_RD_DATA( 1694 downto 1690 ) when CO512_dly2 = "010101101" else 
X_RD_DATA( 1689 downto 1685 ) when CO512_dly2 = "010101110" else 
X_RD_DATA( 1684 downto 1680 ) when CO512_dly2 = "010101111" else 
X_RD_DATA( 1679 downto 1675 ) when CO512_dly2 = "010110000" else 
X_RD_DATA( 1674 downto 1670 ) when CO512_dly2 = "010110001" else 
X_RD_DATA( 1669 downto 1665 ) when CO512_dly2 = "010110010" else 
X_RD_DATA( 1664 downto 1660 ) when CO512_dly2 = "010110011" else 
X_RD_DATA( 1659 downto 1655 ) when CO512_dly2 = "010110100" else 
X_RD_DATA( 1654 downto 1650 ) when CO512_dly2 = "010110101" else 
X_RD_DATA( 1649 downto 1645 ) when CO512_dly2 = "010110110" else 
X_RD_DATA( 1644 downto 1640 ) when CO512_dly2 = "010110111" else 
X_RD_DATA( 1639 downto 1635 ) when CO512_dly2 = "010111000" else 
X_RD_DATA( 1634 downto 1630 ) when CO512_dly2 = "010111001" else 
X_RD_DATA( 1629 downto 1625 ) when CO512_dly2 = "010111010" else 
X_RD_DATA( 1624 downto 1620 ) when CO512_dly2 = "010111011" else 
X_RD_DATA( 1619 downto 1615 ) when CO512_dly2 = "010111100" else 
X_RD_DATA( 1614 downto 1610 ) when CO512_dly2 = "010111101" else 
X_RD_DATA( 1609 downto 1605 ) when CO512_dly2 = "010111110" else 
X_RD_DATA( 1604 downto 1600 ) when CO512_dly2 = "010111111" else 
X_RD_DATA( 1599 downto 1595 ) when CO512_dly2 = "011000000" else 
X_RD_DATA( 1594 downto 1590 ) when CO512_dly2 = "011000001" else 
X_RD_DATA( 1589 downto 1585 ) when CO512_dly2 = "011000010" else 
X_RD_DATA( 1584 downto 1580 ) when CO512_dly2 = "011000011" else 
X_RD_DATA( 1579 downto 1575 ) when CO512_dly2 = "011000100" else 
X_RD_DATA( 1574 downto 1570 ) when CO512_dly2 = "011000101" else 
X_RD_DATA( 1569 downto 1565 ) when CO512_dly2 = "011000110" else 
X_RD_DATA( 1564 downto 1560 ) when CO512_dly2 = "011000111" else 
X_RD_DATA( 1559 downto 1555 ) when CO512_dly2 = "011001000" else 
X_RD_DATA( 1554 downto 1550 ) when CO512_dly2 = "011001001" else 
X_RD_DATA( 1549 downto 1545 ) when CO512_dly2 = "011001010" else 
X_RD_DATA( 1544 downto 1540 ) when CO512_dly2 = "011001011" else 
X_RD_DATA( 1539 downto 1535 ) when CO512_dly2 = "011001100" else 
X_RD_DATA( 1534 downto 1530 ) when CO512_dly2 = "011001101" else 
X_RD_DATA( 1529 downto 1525 ) when CO512_dly2 = "011001110" else 
X_RD_DATA( 1524 downto 1520 ) when CO512_dly2 = "011001111" else 
X_RD_DATA( 1519 downto 1515 ) when CO512_dly2 = "011010000" else 
X_RD_DATA( 1514 downto 1510 ) when CO512_dly2 = "011010001" else 
X_RD_DATA( 1509 downto 1505 ) when CO512_dly2 = "011010010" else 
X_RD_DATA( 1504 downto 1500 ) when CO512_dly2 = "011010011" else 
X_RD_DATA( 1499 downto 1495 ) when CO512_dly2 = "011010100" else 
X_RD_DATA( 1494 downto 1490 ) when CO512_dly2 = "011010101" else 
X_RD_DATA( 1489 downto 1485 ) when CO512_dly2 = "011010110" else 
X_RD_DATA( 1484 downto 1480 ) when CO512_dly2 = "011010111" else 
X_RD_DATA( 1479 downto 1475 ) when CO512_dly2 = "011011000" else 
X_RD_DATA( 1474 downto 1470 ) when CO512_dly2 = "011011001" else 
X_RD_DATA( 1469 downto 1465 ) when CO512_dly2 = "011011010" else 
X_RD_DATA( 1464 downto 1460 ) when CO512_dly2 = "011011011" else 
X_RD_DATA( 1459 downto 1455 ) when CO512_dly2 = "011011100" else 
X_RD_DATA( 1454 downto 1450 ) when CO512_dly2 = "011011101" else 
X_RD_DATA( 1449 downto 1445 ) when CO512_dly2 = "011011110" else 
X_RD_DATA( 1444 downto 1440 ) when CO512_dly2 = "011011111" else 
X_RD_DATA( 1439 downto 1435 ) when CO512_dly2 = "011100000" else 
X_RD_DATA( 1434 downto 1430 ) when CO512_dly2 = "011100001" else 
X_RD_DATA( 1429 downto 1425 ) when CO512_dly2 = "011100010" else 
X_RD_DATA( 1424 downto 1420 ) when CO512_dly2 = "011100011" else 
X_RD_DATA( 1419 downto 1415 ) when CO512_dly2 = "011100100" else 
X_RD_DATA( 1414 downto 1410 ) when CO512_dly2 = "011100101" else 
X_RD_DATA( 1409 downto 1405 ) when CO512_dly2 = "011100110" else 
X_RD_DATA( 1404 downto 1400 ) when CO512_dly2 = "011100111" else 
X_RD_DATA( 1399 downto 1395 ) when CO512_dly2 = "011101000" else 
X_RD_DATA( 1394 downto 1390 ) when CO512_dly2 = "011101001" else 
X_RD_DATA( 1389 downto 1385 ) when CO512_dly2 = "011101010" else 
X_RD_DATA( 1384 downto 1380 ) when CO512_dly2 = "011101011" else 
X_RD_DATA( 1379 downto 1375 ) when CO512_dly2 = "011101100" else 
X_RD_DATA( 1374 downto 1370 ) when CO512_dly2 = "011101101" else 
X_RD_DATA( 1369 downto 1365 ) when CO512_dly2 = "011101110" else 
X_RD_DATA( 1364 downto 1360 ) when CO512_dly2 = "011101111" else 
X_RD_DATA( 1359 downto 1355 ) when CO512_dly2 = "011110000" else 
X_RD_DATA( 1354 downto 1350 ) when CO512_dly2 = "011110001" else 
X_RD_DATA( 1349 downto 1345 ) when CO512_dly2 = "011110010" else 
X_RD_DATA( 1344 downto 1340 ) when CO512_dly2 = "011110011" else 
X_RD_DATA( 1339 downto 1335 ) when CO512_dly2 = "011110100" else 
X_RD_DATA( 1334 downto 1330 ) when CO512_dly2 = "011110101" else 
X_RD_DATA( 1329 downto 1325 ) when CO512_dly2 = "011110110" else 
X_RD_DATA( 1324 downto 1320 ) when CO512_dly2 = "011110111" else 
X_RD_DATA( 1319 downto 1315 ) when CO512_dly2 = "011111000" else 
X_RD_DATA( 1314 downto 1310 ) when CO512_dly2 = "011111001" else 
X_RD_DATA( 1309 downto 1305 ) when CO512_dly2 = "011111010" else 
X_RD_DATA( 1304 downto 1300 ) when CO512_dly2 = "011111011" else 
X_RD_DATA( 1299 downto 1295 ) when CO512_dly2 = "011111100" else 
X_RD_DATA( 1294 downto 1290 ) when CO512_dly2 = "011111101" else 
X_RD_DATA( 1289 downto 1285 ) when CO512_dly2 = "011111110" else 
X_RD_DATA( 1284 downto 1280 ) when CO512_dly2 = "011111111" else 
X_RD_DATA( 1279 downto 1275 ) when CO512_dly2 = "100000000" else 
X_RD_DATA( 1274 downto 1270 ) when CO512_dly2 = "100000001" else 
X_RD_DATA( 1269 downto 1265 ) when CO512_dly2 = "100000010" else 
X_RD_DATA( 1264 downto 1260 ) when CO512_dly2 = "100000011" else 
X_RD_DATA( 1259 downto 1255 ) when CO512_dly2 = "100000100" else 
X_RD_DATA( 1254 downto 1250 ) when CO512_dly2 = "100000101" else 
X_RD_DATA( 1249 downto 1245 ) when CO512_dly2 = "100000110" else 
X_RD_DATA( 1244 downto 1240 ) when CO512_dly2 = "100000111" else 
X_RD_DATA( 1239 downto 1235 ) when CO512_dly2 = "100001000" else 
X_RD_DATA( 1234 downto 1230 ) when CO512_dly2 = "100001001" else 
X_RD_DATA( 1229 downto 1225 ) when CO512_dly2 = "100001010" else 
X_RD_DATA( 1224 downto 1220 ) when CO512_dly2 = "100001011" else 
X_RD_DATA( 1219 downto 1215 ) when CO512_dly2 = "100001100" else 
X_RD_DATA( 1214 downto 1210 ) when CO512_dly2 = "100001101" else 
X_RD_DATA( 1209 downto 1205 ) when CO512_dly2 = "100001110" else 
X_RD_DATA( 1204 downto 1200 ) when CO512_dly2 = "100001111" else 
X_RD_DATA( 1199 downto 1195 ) when CO512_dly2 = "100010000" else 
X_RD_DATA( 1194 downto 1190 ) when CO512_dly2 = "100010001" else 
X_RD_DATA( 1189 downto 1185 ) when CO512_dly2 = "100010010" else 
X_RD_DATA( 1184 downto 1180 ) when CO512_dly2 = "100010011" else 
X_RD_DATA( 1179 downto 1175 ) when CO512_dly2 = "100010100" else 
X_RD_DATA( 1174 downto 1170 ) when CO512_dly2 = "100010101" else 
X_RD_DATA( 1169 downto 1165 ) when CO512_dly2 = "100010110" else 
X_RD_DATA( 1164 downto 1160 ) when CO512_dly2 = "100010111" else 
X_RD_DATA( 1159 downto 1155 ) when CO512_dly2 = "100011000" else 
X_RD_DATA( 1154 downto 1150 ) when CO512_dly2 = "100011001" else 
X_RD_DATA( 1149 downto 1145 ) when CO512_dly2 = "100011010" else 
X_RD_DATA( 1144 downto 1140 ) when CO512_dly2 = "100011011" else 
X_RD_DATA( 1139 downto 1135 ) when CO512_dly2 = "100011100" else 
X_RD_DATA( 1134 downto 1130 ) when CO512_dly2 = "100011101" else 
X_RD_DATA( 1129 downto 1125 ) when CO512_dly2 = "100011110" else 
X_RD_DATA( 1124 downto 1120 ) when CO512_dly2 = "100011111" else 
X_RD_DATA( 1119 downto 1115 ) when CO512_dly2 = "100100000" else 
X_RD_DATA( 1114 downto 1110 ) when CO512_dly2 = "100100001" else 
X_RD_DATA( 1109 downto 1105 ) when CO512_dly2 = "100100010" else 
X_RD_DATA( 1104 downto 1100 ) when CO512_dly2 = "100100011" else 
X_RD_DATA( 1099 downto 1095 ) when CO512_dly2 = "100100100" else 
X_RD_DATA( 1094 downto 1090 ) when CO512_dly2 = "100100101" else 
X_RD_DATA( 1089 downto 1085 ) when CO512_dly2 = "100100110" else 
X_RD_DATA( 1084 downto 1080 ) when CO512_dly2 = "100100111" else 
X_RD_DATA( 1079 downto 1075 ) when CO512_dly2 = "100101000" else 
X_RD_DATA( 1074 downto 1070 ) when CO512_dly2 = "100101001" else 
X_RD_DATA( 1069 downto 1065 ) when CO512_dly2 = "100101010" else 
X_RD_DATA( 1064 downto 1060 ) when CO512_dly2 = "100101011" else 
X_RD_DATA( 1059 downto 1055 ) when CO512_dly2 = "100101100" else 
X_RD_DATA( 1054 downto 1050 ) when CO512_dly2 = "100101101" else 
X_RD_DATA( 1049 downto 1045 ) when CO512_dly2 = "100101110" else 
X_RD_DATA( 1044 downto 1040 ) when CO512_dly2 = "100101111" else 
X_RD_DATA( 1039 downto 1035 ) when CO512_dly2 = "100110000" else 
X_RD_DATA( 1034 downto 1030 ) when CO512_dly2 = "100110001" else 
X_RD_DATA( 1029 downto 1025 ) when CO512_dly2 = "100110010" else 
X_RD_DATA( 1024 downto 1020 ) when CO512_dly2 = "100110011" else 
X_RD_DATA( 1019 downto 1015 ) when CO512_dly2 = "100110100" else 
X_RD_DATA( 1014 downto 1010 ) when CO512_dly2 = "100110101" else 
X_RD_DATA( 1009 downto 1005 ) when CO512_dly2 = "100110110" else 
X_RD_DATA( 1004 downto 1000 ) when CO512_dly2 = "100110111" else 
X_RD_DATA( 999 downto 995 ) when CO512_dly2 = "100111000" else 
X_RD_DATA( 994 downto 990 ) when CO512_dly2 = "100111001" else 
X_RD_DATA( 989 downto 985 ) when CO512_dly2 = "100111010" else 
X_RD_DATA( 984 downto 980 ) when CO512_dly2 = "100111011" else 
X_RD_DATA( 979 downto 975 ) when CO512_dly2 = "100111100" else 
X_RD_DATA( 974 downto 970 ) when CO512_dly2 = "100111101" else 
X_RD_DATA( 969 downto 965 ) when CO512_dly2 = "100111110" else 
X_RD_DATA( 964 downto 960 ) when CO512_dly2 = "100111111" else 
X_RD_DATA( 959 downto 955 ) when CO512_dly2 = "101000000" else 
X_RD_DATA( 954 downto 950 ) when CO512_dly2 = "101000001" else 
X_RD_DATA( 949 downto 945 ) when CO512_dly2 = "101000010" else 
X_RD_DATA( 944 downto 940 ) when CO512_dly2 = "101000011" else 
X_RD_DATA( 939 downto 935 ) when CO512_dly2 = "101000100" else 
X_RD_DATA( 934 downto 930 ) when CO512_dly2 = "101000101" else 
X_RD_DATA( 929 downto 925 ) when CO512_dly2 = "101000110" else 
X_RD_DATA( 924 downto 920 ) when CO512_dly2 = "101000111" else 
X_RD_DATA( 919 downto 915 ) when CO512_dly2 = "101001000" else 
X_RD_DATA( 914 downto 910 ) when CO512_dly2 = "101001001" else 
X_RD_DATA( 909 downto 905 ) when CO512_dly2 = "101001010" else 
X_RD_DATA( 904 downto 900 ) when CO512_dly2 = "101001011" else 
X_RD_DATA( 899 downto 895 ) when CO512_dly2 = "101001100" else 
X_RD_DATA( 894 downto 890 ) when CO512_dly2 = "101001101" else 
X_RD_DATA( 889 downto 885 ) when CO512_dly2 = "101001110" else 
X_RD_DATA( 884 downto 880 ) when CO512_dly2 = "101001111" else 
X_RD_DATA( 879 downto 875 ) when CO512_dly2 = "101010000" else 
X_RD_DATA( 874 downto 870 ) when CO512_dly2 = "101010001" else 
X_RD_DATA( 869 downto 865 ) when CO512_dly2 = "101010010" else 
X_RD_DATA( 864 downto 860 ) when CO512_dly2 = "101010011" else 
X_RD_DATA( 859 downto 855 ) when CO512_dly2 = "101010100" else 
X_RD_DATA( 854 downto 850 ) when CO512_dly2 = "101010101" else 
X_RD_DATA( 849 downto 845 ) when CO512_dly2 = "101010110" else 
X_RD_DATA( 844 downto 840 ) when CO512_dly2 = "101010111" else 
X_RD_DATA( 839 downto 835 ) when CO512_dly2 = "101011000" else 
X_RD_DATA( 834 downto 830 ) when CO512_dly2 = "101011001" else 
X_RD_DATA( 829 downto 825 ) when CO512_dly2 = "101011010" else 
X_RD_DATA( 824 downto 820 ) when CO512_dly2 = "101011011" else 
X_RD_DATA( 819 downto 815 ) when CO512_dly2 = "101011100" else 
X_RD_DATA( 814 downto 810 ) when CO512_dly2 = "101011101" else 
X_RD_DATA( 809 downto 805 ) when CO512_dly2 = "101011110" else 
X_RD_DATA( 804 downto 800 ) when CO512_dly2 = "101011111" else 
X_RD_DATA( 799 downto 795 ) when CO512_dly2 = "101100000" else 
X_RD_DATA( 794 downto 790 ) when CO512_dly2 = "101100001" else 
X_RD_DATA( 789 downto 785 ) when CO512_dly2 = "101100010" else 
X_RD_DATA( 784 downto 780 ) when CO512_dly2 = "101100011" else 
X_RD_DATA( 779 downto 775 ) when CO512_dly2 = "101100100" else 
X_RD_DATA( 774 downto 770 ) when CO512_dly2 = "101100101" else 
X_RD_DATA( 769 downto 765 ) when CO512_dly2 = "101100110" else 
X_RD_DATA( 764 downto 760 ) when CO512_dly2 = "101100111" else 
X_RD_DATA( 759 downto 755 ) when CO512_dly2 = "101101000" else 
X_RD_DATA( 754 downto 750 ) when CO512_dly2 = "101101001" else 
X_RD_DATA( 749 downto 745 ) when CO512_dly2 = "101101010" else 
X_RD_DATA( 744 downto 740 ) when CO512_dly2 = "101101011" else 
X_RD_DATA( 739 downto 735 ) when CO512_dly2 = "101101100" else 
X_RD_DATA( 734 downto 730 ) when CO512_dly2 = "101101101" else 
X_RD_DATA( 729 downto 725 ) when CO512_dly2 = "101101110" else 
X_RD_DATA( 724 downto 720 ) when CO512_dly2 = "101101111" else 
X_RD_DATA( 719 downto 715 ) when CO512_dly2 = "101110000" else 
X_RD_DATA( 714 downto 710 ) when CO512_dly2 = "101110001" else 
X_RD_DATA( 709 downto 705 ) when CO512_dly2 = "101110010" else 
X_RD_DATA( 704 downto 700 ) when CO512_dly2 = "101110011" else 
X_RD_DATA( 699 downto 695 ) when CO512_dly2 = "101110100" else 
X_RD_DATA( 694 downto 690 ) when CO512_dly2 = "101110101" else 
X_RD_DATA( 689 downto 685 ) when CO512_dly2 = "101110110" else 
X_RD_DATA( 684 downto 680 ) when CO512_dly2 = "101110111" else 
X_RD_DATA( 679 downto 675 ) when CO512_dly2 = "101111000" else 
X_RD_DATA( 674 downto 670 ) when CO512_dly2 = "101111001" else 
X_RD_DATA( 669 downto 665 ) when CO512_dly2 = "101111010" else 
X_RD_DATA( 664 downto 660 ) when CO512_dly2 = "101111011" else 
X_RD_DATA( 659 downto 655 ) when CO512_dly2 = "101111100" else 
X_RD_DATA( 654 downto 650 ) when CO512_dly2 = "101111101" else 
X_RD_DATA( 649 downto 645 ) when CO512_dly2 = "101111110" else 
X_RD_DATA( 644 downto 640 ) when CO512_dly2 = "101111111" else 
X_RD_DATA( 639 downto 635 ) when CO512_dly2 = "110000000" else 
X_RD_DATA( 634 downto 630 ) when CO512_dly2 = "110000001" else 
X_RD_DATA( 629 downto 625 ) when CO512_dly2 = "110000010" else 
X_RD_DATA( 624 downto 620 ) when CO512_dly2 = "110000011" else 
X_RD_DATA( 619 downto 615 ) when CO512_dly2 = "110000100" else 
X_RD_DATA( 614 downto 610 ) when CO512_dly2 = "110000101" else 
X_RD_DATA( 609 downto 605 ) when CO512_dly2 = "110000110" else 
X_RD_DATA( 604 downto 600 ) when CO512_dly2 = "110000111" else 
X_RD_DATA( 599 downto 595 ) when CO512_dly2 = "110001000" else 
X_RD_DATA( 594 downto 590 ) when CO512_dly2 = "110001001" else 
X_RD_DATA( 589 downto 585 ) when CO512_dly2 = "110001010" else 
X_RD_DATA( 584 downto 580 ) when CO512_dly2 = "110001011" else 
X_RD_DATA( 579 downto 575 ) when CO512_dly2 = "110001100" else 
X_RD_DATA( 574 downto 570 ) when CO512_dly2 = "110001101" else 
X_RD_DATA( 569 downto 565 ) when CO512_dly2 = "110001110" else 
X_RD_DATA( 564 downto 560 ) when CO512_dly2 = "110001111" else 
X_RD_DATA( 559 downto 555 ) when CO512_dly2 = "110010000" else 
X_RD_DATA( 554 downto 550 ) when CO512_dly2 = "110010001" else 
X_RD_DATA( 549 downto 545 ) when CO512_dly2 = "110010010" else 
X_RD_DATA( 544 downto 540 ) when CO512_dly2 = "110010011" else 
X_RD_DATA( 539 downto 535 ) when CO512_dly2 = "110010100" else 
X_RD_DATA( 534 downto 530 ) when CO512_dly2 = "110010101" else 
X_RD_DATA( 529 downto 525 ) when CO512_dly2 = "110010110" else 
X_RD_DATA( 524 downto 520 ) when CO512_dly2 = "110010111" else 
X_RD_DATA( 519 downto 515 ) when CO512_dly2 = "110011000" else 
X_RD_DATA( 514 downto 510 ) when CO512_dly2 = "110011001" else 
X_RD_DATA( 509 downto 505 ) when CO512_dly2 = "110011010" else 
X_RD_DATA( 504 downto 500 ) when CO512_dly2 = "110011011" else 
X_RD_DATA( 499 downto 495 ) when CO512_dly2 = "110011100" else 
X_RD_DATA( 494 downto 490 ) when CO512_dly2 = "110011101" else 
X_RD_DATA( 489 downto 485 ) when CO512_dly2 = "110011110" else 
X_RD_DATA( 484 downto 480 ) when CO512_dly2 = "110011111" else 
X_RD_DATA( 479 downto 475 ) when CO512_dly2 = "110100000" else 
X_RD_DATA( 474 downto 470 ) when CO512_dly2 = "110100001" else 
X_RD_DATA( 469 downto 465 ) when CO512_dly2 = "110100010" else 
X_RD_DATA( 464 downto 460 ) when CO512_dly2 = "110100011" else 
X_RD_DATA( 459 downto 455 ) when CO512_dly2 = "110100100" else 
X_RD_DATA( 454 downto 450 ) when CO512_dly2 = "110100101" else 
X_RD_DATA( 449 downto 445 ) when CO512_dly2 = "110100110" else 
X_RD_DATA( 444 downto 440 ) when CO512_dly2 = "110100111" else 
X_RD_DATA( 439 downto 435 ) when CO512_dly2 = "110101000" else 
X_RD_DATA( 434 downto 430 ) when CO512_dly2 = "110101001" else 
X_RD_DATA( 429 downto 425 ) when CO512_dly2 = "110101010" else 
X_RD_DATA( 424 downto 420 ) when CO512_dly2 = "110101011" else 
X_RD_DATA( 419 downto 415 ) when CO512_dly2 = "110101100" else 
X_RD_DATA( 414 downto 410 ) when CO512_dly2 = "110101101" else 
X_RD_DATA( 409 downto 405 ) when CO512_dly2 = "110101110" else 
X_RD_DATA( 404 downto 400 ) when CO512_dly2 = "110101111" else 
X_RD_DATA( 399 downto 395 ) when CO512_dly2 = "110110000" else 
X_RD_DATA( 394 downto 390 ) when CO512_dly2 = "110110001" else 
X_RD_DATA( 389 downto 385 ) when CO512_dly2 = "110110010" else 
X_RD_DATA( 384 downto 380 ) when CO512_dly2 = "110110011" else 
X_RD_DATA( 379 downto 375 ) when CO512_dly2 = "110110100" else 
X_RD_DATA( 374 downto 370 ) when CO512_dly2 = "110110101" else 
X_RD_DATA( 369 downto 365 ) when CO512_dly2 = "110110110" else 
X_RD_DATA( 364 downto 360 ) when CO512_dly2 = "110110111" else 
X_RD_DATA( 359 downto 355 ) when CO512_dly2 = "110111000" else 
X_RD_DATA( 354 downto 350 ) when CO512_dly2 = "110111001" else 
X_RD_DATA( 349 downto 345 ) when CO512_dly2 = "110111010" else 
X_RD_DATA( 344 downto 340 ) when CO512_dly2 = "110111011" else 
X_RD_DATA( 339 downto 335 ) when CO512_dly2 = "110111100" else 
X_RD_DATA( 334 downto 330 ) when CO512_dly2 = "110111101" else 
X_RD_DATA( 329 downto 325 ) when CO512_dly2 = "110111110" else 
X_RD_DATA( 324 downto 320 ) when CO512_dly2 = "110111111" else 
X_RD_DATA( 319 downto 315 ) when CO512_dly2 = "111000000" else 
X_RD_DATA( 314 downto 310 ) when CO512_dly2 = "111000001" else 
X_RD_DATA( 309 downto 305 ) when CO512_dly2 = "111000010" else 
X_RD_DATA( 304 downto 300 ) when CO512_dly2 = "111000011" else 
X_RD_DATA( 299 downto 295 ) when CO512_dly2 = "111000100" else 
X_RD_DATA( 294 downto 290 ) when CO512_dly2 = "111000101" else 
X_RD_DATA( 289 downto 285 ) when CO512_dly2 = "111000110" else 
X_RD_DATA( 284 downto 280 ) when CO512_dly2 = "111000111" else 
X_RD_DATA( 279 downto 275 ) when CO512_dly2 = "111001000" else 
X_RD_DATA( 274 downto 270 ) when CO512_dly2 = "111001001" else 
X_RD_DATA( 269 downto 265 ) when CO512_dly2 = "111001010" else 
X_RD_DATA( 264 downto 260 ) when CO512_dly2 = "111001011" else 
X_RD_DATA( 259 downto 255 ) when CO512_dly2 = "111001100" else 
X_RD_DATA( 254 downto 250 ) when CO512_dly2 = "111001101" else 
X_RD_DATA( 249 downto 245 ) when CO512_dly2 = "111001110" else 
X_RD_DATA( 244 downto 240 ) when CO512_dly2 = "111001111" else 
X_RD_DATA( 239 downto 235 ) when CO512_dly2 = "111010000" else 
X_RD_DATA( 234 downto 230 ) when CO512_dly2 = "111010001" else 
X_RD_DATA( 229 downto 225 ) when CO512_dly2 = "111010010" else 
X_RD_DATA( 224 downto 220 ) when CO512_dly2 = "111010011" else 
X_RD_DATA( 219 downto 215 ) when CO512_dly2 = "111010100" else 
X_RD_DATA( 214 downto 210 ) when CO512_dly2 = "111010101" else 
X_RD_DATA( 209 downto 205 ) when CO512_dly2 = "111010110" else 
X_RD_DATA( 204 downto 200 ) when CO512_dly2 = "111010111" else 
X_RD_DATA( 199 downto 195 ) when CO512_dly2 = "111011000" else 
X_RD_DATA( 194 downto 190 ) when CO512_dly2 = "111011001" else 
X_RD_DATA( 189 downto 185 ) when CO512_dly2 = "111011010" else 
X_RD_DATA( 184 downto 180 ) when CO512_dly2 = "111011011" else 
X_RD_DATA( 179 downto 175 ) when CO512_dly2 = "111011100" else 
X_RD_DATA( 174 downto 170 ) when CO512_dly2 = "111011101" else 
X_RD_DATA( 169 downto 165 ) when CO512_dly2 = "111011110" else 
X_RD_DATA( 164 downto 160 ) when CO512_dly2 = "111011111" else 
X_RD_DATA( 159 downto 155 ) when CO512_dly2 = "111100000" else 
X_RD_DATA( 154 downto 150 ) when CO512_dly2 = "111100001" else 
X_RD_DATA( 149 downto 145 ) when CO512_dly2 = "111100010" else 
X_RD_DATA( 144 downto 140 ) when CO512_dly2 = "111100011" else 
X_RD_DATA( 139 downto 135 ) when CO512_dly2 = "111100100" else 
X_RD_DATA( 134 downto 130 ) when CO512_dly2 = "111100101" else 
X_RD_DATA( 129 downto 125 ) when CO512_dly2 = "111100110" else 
X_RD_DATA( 124 downto 120 ) when CO512_dly2 = "111100111" else 
X_RD_DATA( 119 downto 115 ) when CO512_dly2 = "111101000" else 
X_RD_DATA( 114 downto 110 ) when CO512_dly2 = "111101001" else 
X_RD_DATA( 109 downto 105 ) when CO512_dly2 = "111101010" else 
X_RD_DATA( 104 downto 100 ) when CO512_dly2 = "111101011" else 
X_RD_DATA( 99 downto 95 ) when CO512_dly2 = "111101100" else 
X_RD_DATA( 94 downto 90 ) when CO512_dly2 = "111101101" else 
X_RD_DATA( 89 downto 85 ) when CO512_dly2 = "111101110" else 
X_RD_DATA( 84 downto 80 ) when CO512_dly2 = "111101111" else 
X_RD_DATA( 79 downto 75 ) when CO512_dly2 = "111110000" else 
X_RD_DATA( 74 downto 70 ) when CO512_dly2 = "111110001" else 
X_RD_DATA( 69 downto 65 ) when CO512_dly2 = "111110010" else 
X_RD_DATA( 64 downto 60 ) when CO512_dly2 = "111110011" else 
X_RD_DATA( 59 downto 55 ) when CO512_dly2 = "111110100" else 
X_RD_DATA( 54 downto 50 ) when CO512_dly2 = "111110101" else 
X_RD_DATA( 49 downto 45 ) when CO512_dly2 = "111110110" else 
X_RD_DATA( 44 downto 40 ) when CO512_dly2 = "111110111" else 
X_RD_DATA( 39 downto 35 ) when CO512_dly2 = "111111000" else 
X_RD_DATA( 34 downto 30 ) when CO512_dly2 = "111111001" else 
X_RD_DATA( 29 downto 25 ) when CO512_dly2 = "111111010" else 
X_RD_DATA( 24 downto 20 ) when CO512_dly2 = "111111011" else 
X_RD_DATA( 19 downto 15 ) when CO512_dly2 = "111111100" else 
X_RD_DATA( 14 downto 10 ) when CO512_dly2 = "111111101" else 
X_RD_DATA( 9 downto 5 ) when CO512_dly2 = "111111110" else 
X_RD_DATA( 4 downto 0 ); -- when CO512_dly2 = "111111111" else 


W9Addr_L <= CO9;
W9Addr_H <= ("0" & CO512 & "000") + ("0000" & CO512); -- CO512 x 9
W9Addr_prev <= W9Addr_H + ("00000000" & W9Addr_L);

---------------------------------------------------------
process( RST_N, CLK )  
begin 
	if( RST_N = '0' ) then
		W9Addr <= (others => '0');
	elsif( CLK'event and CLK = '1' ) then
		if(W9Addr_prev >= -448 and W9Addr_prev <= -1) then -- note that this is Signed
			W9Addr <= (others => '0');
		else
			W9Addr <= W9Addr_prev;
		end if;
	end if;  
end process;

---------------------------------------------------------
enb_a <= '1' when W9Sel = 1 and CO512 = 511 and CO9 = 8 else '0';
enb_b <= '1' when W9Sel = 2 and CO512 = 511 and CO9 = 8 else '0';
enb_c <= '1' when W9Sel = 3 and CO512 = 511 and CO9 = 8 else '0';
enb_d <= '1' when W9Sel = 0 and CO512 = 511 and CO9 = 8 else '0';

-----------------------------------------------------------
process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_000_0 <= Relu_000;
		elsif(enb_b = '1') then
			reg_000_1 <= Relu_000;
		elsif(enb_c = '1') then
			reg_000_2 <= Relu_000;
		elsif(enb_d = '1') then
			reg_000_3 <= Relu_000;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_001_0 <= Relu_001;
		elsif(enb_b = '1') then
			reg_001_1 <= Relu_001;
		elsif(enb_c = '1') then
			reg_001_2 <= Relu_001;
		elsif(enb_d = '1') then
			reg_001_3 <= Relu_001;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_002_0 <= Relu_002;
		elsif(enb_b = '1') then
			reg_002_1 <= Relu_002;
		elsif(enb_c = '1') then
			reg_002_2 <= Relu_002;
		elsif(enb_d = '1') then
			reg_002_3 <= Relu_002;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_003_0 <= Relu_003;
		elsif(enb_b = '1') then
			reg_003_1 <= Relu_003;
		elsif(enb_c = '1') then
			reg_003_2 <= Relu_003;
		elsif(enb_d = '1') then
			reg_003_3 <= Relu_003;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_004_0 <= Relu_004;
		elsif(enb_b = '1') then
			reg_004_1 <= Relu_004;
		elsif(enb_c = '1') then
			reg_004_2 <= Relu_004;
		elsif(enb_d = '1') then
			reg_004_3 <= Relu_004;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_005_0 <= Relu_005;
		elsif(enb_b = '1') then
			reg_005_1 <= Relu_005;
		elsif(enb_c = '1') then
			reg_005_2 <= Relu_005;
		elsif(enb_d = '1') then
			reg_005_3 <= Relu_005;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_006_0 <= Relu_006;
		elsif(enb_b = '1') then
			reg_006_1 <= Relu_006;
		elsif(enb_c = '1') then
			reg_006_2 <= Relu_006;
		elsif(enb_d = '1') then
			reg_006_3 <= Relu_006;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_007_0 <= Relu_007;
		elsif(enb_b = '1') then
			reg_007_1 <= Relu_007;
		elsif(enb_c = '1') then
			reg_007_2 <= Relu_007;
		elsif(enb_d = '1') then
			reg_007_3 <= Relu_007;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_008_0 <= Relu_008;
		elsif(enb_b = '1') then
			reg_008_1 <= Relu_008;
		elsif(enb_c = '1') then
			reg_008_2 <= Relu_008;
		elsif(enb_d = '1') then
			reg_008_3 <= Relu_008;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_009_0 <= Relu_009;
		elsif(enb_b = '1') then
			reg_009_1 <= Relu_009;
		elsif(enb_c = '1') then
			reg_009_2 <= Relu_009;
		elsif(enb_d = '1') then
			reg_009_3 <= Relu_009;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_010_0 <= Relu_010;
		elsif(enb_b = '1') then
			reg_010_1 <= Relu_010;
		elsif(enb_c = '1') then
			reg_010_2 <= Relu_010;
		elsif(enb_d = '1') then
			reg_010_3 <= Relu_010;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_011_0 <= Relu_011;
		elsif(enb_b = '1') then
			reg_011_1 <= Relu_011;
		elsif(enb_c = '1') then
			reg_011_2 <= Relu_011;
		elsif(enb_d = '1') then
			reg_011_3 <= Relu_011;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_012_0 <= Relu_012;
		elsif(enb_b = '1') then
			reg_012_1 <= Relu_012;
		elsif(enb_c = '1') then
			reg_012_2 <= Relu_012;
		elsif(enb_d = '1') then
			reg_012_3 <= Relu_012;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_013_0 <= Relu_013;
		elsif(enb_b = '1') then
			reg_013_1 <= Relu_013;
		elsif(enb_c = '1') then
			reg_013_2 <= Relu_013;
		elsif(enb_d = '1') then
			reg_013_3 <= Relu_013;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_014_0 <= Relu_014;
		elsif(enb_b = '1') then
			reg_014_1 <= Relu_014;
		elsif(enb_c = '1') then
			reg_014_2 <= Relu_014;
		elsif(enb_d = '1') then
			reg_014_3 <= Relu_014;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_015_0 <= Relu_015;
		elsif(enb_b = '1') then
			reg_015_1 <= Relu_015;
		elsif(enb_c = '1') then
			reg_015_2 <= Relu_015;
		elsif(enb_d = '1') then
			reg_015_3 <= Relu_015;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_016_0 <= Relu_016;
		elsif(enb_b = '1') then
			reg_016_1 <= Relu_016;
		elsif(enb_c = '1') then
			reg_016_2 <= Relu_016;
		elsif(enb_d = '1') then
			reg_016_3 <= Relu_016;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_017_0 <= Relu_017;
		elsif(enb_b = '1') then
			reg_017_1 <= Relu_017;
		elsif(enb_c = '1') then
			reg_017_2 <= Relu_017;
		elsif(enb_d = '1') then
			reg_017_3 <= Relu_017;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_018_0 <= Relu_018;
		elsif(enb_b = '1') then
			reg_018_1 <= Relu_018;
		elsif(enb_c = '1') then
			reg_018_2 <= Relu_018;
		elsif(enb_d = '1') then
			reg_018_3 <= Relu_018;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_019_0 <= Relu_019;
		elsif(enb_b = '1') then
			reg_019_1 <= Relu_019;
		elsif(enb_c = '1') then
			reg_019_2 <= Relu_019;
		elsif(enb_d = '1') then
			reg_019_3 <= Relu_019;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_020_0 <= Relu_020;
		elsif(enb_b = '1') then
			reg_020_1 <= Relu_020;
		elsif(enb_c = '1') then
			reg_020_2 <= Relu_020;
		elsif(enb_d = '1') then
			reg_020_3 <= Relu_020;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_021_0 <= Relu_021;
		elsif(enb_b = '1') then
			reg_021_1 <= Relu_021;
		elsif(enb_c = '1') then
			reg_021_2 <= Relu_021;
		elsif(enb_d = '1') then
			reg_021_3 <= Relu_021;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_022_0 <= Relu_022;
		elsif(enb_b = '1') then
			reg_022_1 <= Relu_022;
		elsif(enb_c = '1') then
			reg_022_2 <= Relu_022;
		elsif(enb_d = '1') then
			reg_022_3 <= Relu_022;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_023_0 <= Relu_023;
		elsif(enb_b = '1') then
			reg_023_1 <= Relu_023;
		elsif(enb_c = '1') then
			reg_023_2 <= Relu_023;
		elsif(enb_d = '1') then
			reg_023_3 <= Relu_023;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_024_0 <= Relu_024;
		elsif(enb_b = '1') then
			reg_024_1 <= Relu_024;
		elsif(enb_c = '1') then
			reg_024_2 <= Relu_024;
		elsif(enb_d = '1') then
			reg_024_3 <= Relu_024;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_025_0 <= Relu_025;
		elsif(enb_b = '1') then
			reg_025_1 <= Relu_025;
		elsif(enb_c = '1') then
			reg_025_2 <= Relu_025;
		elsif(enb_d = '1') then
			reg_025_3 <= Relu_025;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_026_0 <= Relu_026;
		elsif(enb_b = '1') then
			reg_026_1 <= Relu_026;
		elsif(enb_c = '1') then
			reg_026_2 <= Relu_026;
		elsif(enb_d = '1') then
			reg_026_3 <= Relu_026;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_027_0 <= Relu_027;
		elsif(enb_b = '1') then
			reg_027_1 <= Relu_027;
		elsif(enb_c = '1') then
			reg_027_2 <= Relu_027;
		elsif(enb_d = '1') then
			reg_027_3 <= Relu_027;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_028_0 <= Relu_028;
		elsif(enb_b = '1') then
			reg_028_1 <= Relu_028;
		elsif(enb_c = '1') then
			reg_028_2 <= Relu_028;
		elsif(enb_d = '1') then
			reg_028_3 <= Relu_028;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_029_0 <= Relu_029;
		elsif(enb_b = '1') then
			reg_029_1 <= Relu_029;
		elsif(enb_c = '1') then
			reg_029_2 <= Relu_029;
		elsif(enb_d = '1') then
			reg_029_3 <= Relu_029;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_030_0 <= Relu_030;
		elsif(enb_b = '1') then
			reg_030_1 <= Relu_030;
		elsif(enb_c = '1') then
			reg_030_2 <= Relu_030;
		elsif(enb_d = '1') then
			reg_030_3 <= Relu_030;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_031_0 <= Relu_031;
		elsif(enb_b = '1') then
			reg_031_1 <= Relu_031;
		elsif(enb_c = '1') then
			reg_031_2 <= Relu_031;
		elsif(enb_d = '1') then
			reg_031_3 <= Relu_031;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_032_0 <= Relu_032;
		elsif(enb_b = '1') then
			reg_032_1 <= Relu_032;
		elsif(enb_c = '1') then
			reg_032_2 <= Relu_032;
		elsif(enb_d = '1') then
			reg_032_3 <= Relu_032;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_033_0 <= Relu_033;
		elsif(enb_b = '1') then
			reg_033_1 <= Relu_033;
		elsif(enb_c = '1') then
			reg_033_2 <= Relu_033;
		elsif(enb_d = '1') then
			reg_033_3 <= Relu_033;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_034_0 <= Relu_034;
		elsif(enb_b = '1') then
			reg_034_1 <= Relu_034;
		elsif(enb_c = '1') then
			reg_034_2 <= Relu_034;
		elsif(enb_d = '1') then
			reg_034_3 <= Relu_034;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_035_0 <= Relu_035;
		elsif(enb_b = '1') then
			reg_035_1 <= Relu_035;
		elsif(enb_c = '1') then
			reg_035_2 <= Relu_035;
		elsif(enb_d = '1') then
			reg_035_3 <= Relu_035;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_036_0 <= Relu_036;
		elsif(enb_b = '1') then
			reg_036_1 <= Relu_036;
		elsif(enb_c = '1') then
			reg_036_2 <= Relu_036;
		elsif(enb_d = '1') then
			reg_036_3 <= Relu_036;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_037_0 <= Relu_037;
		elsif(enb_b = '1') then
			reg_037_1 <= Relu_037;
		elsif(enb_c = '1') then
			reg_037_2 <= Relu_037;
		elsif(enb_d = '1') then
			reg_037_3 <= Relu_037;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_038_0 <= Relu_038;
		elsif(enb_b = '1') then
			reg_038_1 <= Relu_038;
		elsif(enb_c = '1') then
			reg_038_2 <= Relu_038;
		elsif(enb_d = '1') then
			reg_038_3 <= Relu_038;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_039_0 <= Relu_039;
		elsif(enb_b = '1') then
			reg_039_1 <= Relu_039;
		elsif(enb_c = '1') then
			reg_039_2 <= Relu_039;
		elsif(enb_d = '1') then
			reg_039_3 <= Relu_039;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_040_0 <= Relu_040;
		elsif(enb_b = '1') then
			reg_040_1 <= Relu_040;
		elsif(enb_c = '1') then
			reg_040_2 <= Relu_040;
		elsif(enb_d = '1') then
			reg_040_3 <= Relu_040;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_041_0 <= Relu_041;
		elsif(enb_b = '1') then
			reg_041_1 <= Relu_041;
		elsif(enb_c = '1') then
			reg_041_2 <= Relu_041;
		elsif(enb_d = '1') then
			reg_041_3 <= Relu_041;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_042_0 <= Relu_042;
		elsif(enb_b = '1') then
			reg_042_1 <= Relu_042;
		elsif(enb_c = '1') then
			reg_042_2 <= Relu_042;
		elsif(enb_d = '1') then
			reg_042_3 <= Relu_042;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_043_0 <= Relu_043;
		elsif(enb_b = '1') then
			reg_043_1 <= Relu_043;
		elsif(enb_c = '1') then
			reg_043_2 <= Relu_043;
		elsif(enb_d = '1') then
			reg_043_3 <= Relu_043;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_044_0 <= Relu_044;
		elsif(enb_b = '1') then
			reg_044_1 <= Relu_044;
		elsif(enb_c = '1') then
			reg_044_2 <= Relu_044;
		elsif(enb_d = '1') then
			reg_044_3 <= Relu_044;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_045_0 <= Relu_045;
		elsif(enb_b = '1') then
			reg_045_1 <= Relu_045;
		elsif(enb_c = '1') then
			reg_045_2 <= Relu_045;
		elsif(enb_d = '1') then
			reg_045_3 <= Relu_045;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_046_0 <= Relu_046;
		elsif(enb_b = '1') then
			reg_046_1 <= Relu_046;
		elsif(enb_c = '1') then
			reg_046_2 <= Relu_046;
		elsif(enb_d = '1') then
			reg_046_3 <= Relu_046;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_047_0 <= Relu_047;
		elsif(enb_b = '1') then
			reg_047_1 <= Relu_047;
		elsif(enb_c = '1') then
			reg_047_2 <= Relu_047;
		elsif(enb_d = '1') then
			reg_047_3 <= Relu_047;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_048_0 <= Relu_048;
		elsif(enb_b = '1') then
			reg_048_1 <= Relu_048;
		elsif(enb_c = '1') then
			reg_048_2 <= Relu_048;
		elsif(enb_d = '1') then
			reg_048_3 <= Relu_048;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_049_0 <= Relu_049;
		elsif(enb_b = '1') then
			reg_049_1 <= Relu_049;
		elsif(enb_c = '1') then
			reg_049_2 <= Relu_049;
		elsif(enb_d = '1') then
			reg_049_3 <= Relu_049;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_050_0 <= Relu_050;
		elsif(enb_b = '1') then
			reg_050_1 <= Relu_050;
		elsif(enb_c = '1') then
			reg_050_2 <= Relu_050;
		elsif(enb_d = '1') then
			reg_050_3 <= Relu_050;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_051_0 <= Relu_051;
		elsif(enb_b = '1') then
			reg_051_1 <= Relu_051;
		elsif(enb_c = '1') then
			reg_051_2 <= Relu_051;
		elsif(enb_d = '1') then
			reg_051_3 <= Relu_051;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_052_0 <= Relu_052;
		elsif(enb_b = '1') then
			reg_052_1 <= Relu_052;
		elsif(enb_c = '1') then
			reg_052_2 <= Relu_052;
		elsif(enb_d = '1') then
			reg_052_3 <= Relu_052;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_053_0 <= Relu_053;
		elsif(enb_b = '1') then
			reg_053_1 <= Relu_053;
		elsif(enb_c = '1') then
			reg_053_2 <= Relu_053;
		elsif(enb_d = '1') then
			reg_053_3 <= Relu_053;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_054_0 <= Relu_054;
		elsif(enb_b = '1') then
			reg_054_1 <= Relu_054;
		elsif(enb_c = '1') then
			reg_054_2 <= Relu_054;
		elsif(enb_d = '1') then
			reg_054_3 <= Relu_054;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_055_0 <= Relu_055;
		elsif(enb_b = '1') then
			reg_055_1 <= Relu_055;
		elsif(enb_c = '1') then
			reg_055_2 <= Relu_055;
		elsif(enb_d = '1') then
			reg_055_3 <= Relu_055;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_056_0 <= Relu_056;
		elsif(enb_b = '1') then
			reg_056_1 <= Relu_056;
		elsif(enb_c = '1') then
			reg_056_2 <= Relu_056;
		elsif(enb_d = '1') then
			reg_056_3 <= Relu_056;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_057_0 <= Relu_057;
		elsif(enb_b = '1') then
			reg_057_1 <= Relu_057;
		elsif(enb_c = '1') then
			reg_057_2 <= Relu_057;
		elsif(enb_d = '1') then
			reg_057_3 <= Relu_057;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_058_0 <= Relu_058;
		elsif(enb_b = '1') then
			reg_058_1 <= Relu_058;
		elsif(enb_c = '1') then
			reg_058_2 <= Relu_058;
		elsif(enb_d = '1') then
			reg_058_3 <= Relu_058;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_059_0 <= Relu_059;
		elsif(enb_b = '1') then
			reg_059_1 <= Relu_059;
		elsif(enb_c = '1') then
			reg_059_2 <= Relu_059;
		elsif(enb_d = '1') then
			reg_059_3 <= Relu_059;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_060_0 <= Relu_060;
		elsif(enb_b = '1') then
			reg_060_1 <= Relu_060;
		elsif(enb_c = '1') then
			reg_060_2 <= Relu_060;
		elsif(enb_d = '1') then
			reg_060_3 <= Relu_060;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_061_0 <= Relu_061;
		elsif(enb_b = '1') then
			reg_061_1 <= Relu_061;
		elsif(enb_c = '1') then
			reg_061_2 <= Relu_061;
		elsif(enb_d = '1') then
			reg_061_3 <= Relu_061;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_062_0 <= Relu_062;
		elsif(enb_b = '1') then
			reg_062_1 <= Relu_062;
		elsif(enb_c = '1') then
			reg_062_2 <= Relu_062;
		elsif(enb_d = '1') then
			reg_062_3 <= Relu_062;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_063_0 <= Relu_063;
		elsif(enb_b = '1') then
			reg_063_1 <= Relu_063;
		elsif(enb_c = '1') then
			reg_063_2 <= Relu_063;
		elsif(enb_d = '1') then
			reg_063_3 <= Relu_063;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_064_0 <= Relu_064;
		elsif(enb_b = '1') then
			reg_064_1 <= Relu_064;
		elsif(enb_c = '1') then
			reg_064_2 <= Relu_064;
		elsif(enb_d = '1') then
			reg_064_3 <= Relu_064;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_065_0 <= Relu_065;
		elsif(enb_b = '1') then
			reg_065_1 <= Relu_065;
		elsif(enb_c = '1') then
			reg_065_2 <= Relu_065;
		elsif(enb_d = '1') then
			reg_065_3 <= Relu_065;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_066_0 <= Relu_066;
		elsif(enb_b = '1') then
			reg_066_1 <= Relu_066;
		elsif(enb_c = '1') then
			reg_066_2 <= Relu_066;
		elsif(enb_d = '1') then
			reg_066_3 <= Relu_066;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_067_0 <= Relu_067;
		elsif(enb_b = '1') then
			reg_067_1 <= Relu_067;
		elsif(enb_c = '1') then
			reg_067_2 <= Relu_067;
		elsif(enb_d = '1') then
			reg_067_3 <= Relu_067;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_068_0 <= Relu_068;
		elsif(enb_b = '1') then
			reg_068_1 <= Relu_068;
		elsif(enb_c = '1') then
			reg_068_2 <= Relu_068;
		elsif(enb_d = '1') then
			reg_068_3 <= Relu_068;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_069_0 <= Relu_069;
		elsif(enb_b = '1') then
			reg_069_1 <= Relu_069;
		elsif(enb_c = '1') then
			reg_069_2 <= Relu_069;
		elsif(enb_d = '1') then
			reg_069_3 <= Relu_069;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_070_0 <= Relu_070;
		elsif(enb_b = '1') then
			reg_070_1 <= Relu_070;
		elsif(enb_c = '1') then
			reg_070_2 <= Relu_070;
		elsif(enb_d = '1') then
			reg_070_3 <= Relu_070;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_071_0 <= Relu_071;
		elsif(enb_b = '1') then
			reg_071_1 <= Relu_071;
		elsif(enb_c = '1') then
			reg_071_2 <= Relu_071;
		elsif(enb_d = '1') then
			reg_071_3 <= Relu_071;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_072_0 <= Relu_072;
		elsif(enb_b = '1') then
			reg_072_1 <= Relu_072;
		elsif(enb_c = '1') then
			reg_072_2 <= Relu_072;
		elsif(enb_d = '1') then
			reg_072_3 <= Relu_072;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_073_0 <= Relu_073;
		elsif(enb_b = '1') then
			reg_073_1 <= Relu_073;
		elsif(enb_c = '1') then
			reg_073_2 <= Relu_073;
		elsif(enb_d = '1') then
			reg_073_3 <= Relu_073;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_074_0 <= Relu_074;
		elsif(enb_b = '1') then
			reg_074_1 <= Relu_074;
		elsif(enb_c = '1') then
			reg_074_2 <= Relu_074;
		elsif(enb_d = '1') then
			reg_074_3 <= Relu_074;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_075_0 <= Relu_075;
		elsif(enb_b = '1') then
			reg_075_1 <= Relu_075;
		elsif(enb_c = '1') then
			reg_075_2 <= Relu_075;
		elsif(enb_d = '1') then
			reg_075_3 <= Relu_075;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_076_0 <= Relu_076;
		elsif(enb_b = '1') then
			reg_076_1 <= Relu_076;
		elsif(enb_c = '1') then
			reg_076_2 <= Relu_076;
		elsif(enb_d = '1') then
			reg_076_3 <= Relu_076;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_077_0 <= Relu_077;
		elsif(enb_b = '1') then
			reg_077_1 <= Relu_077;
		elsif(enb_c = '1') then
			reg_077_2 <= Relu_077;
		elsif(enb_d = '1') then
			reg_077_3 <= Relu_077;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_078_0 <= Relu_078;
		elsif(enb_b = '1') then
			reg_078_1 <= Relu_078;
		elsif(enb_c = '1') then
			reg_078_2 <= Relu_078;
		elsif(enb_d = '1') then
			reg_078_3 <= Relu_078;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_079_0 <= Relu_079;
		elsif(enb_b = '1') then
			reg_079_1 <= Relu_079;
		elsif(enb_c = '1') then
			reg_079_2 <= Relu_079;
		elsif(enb_d = '1') then
			reg_079_3 <= Relu_079;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_080_0 <= Relu_080;
		elsif(enb_b = '1') then
			reg_080_1 <= Relu_080;
		elsif(enb_c = '1') then
			reg_080_2 <= Relu_080;
		elsif(enb_d = '1') then
			reg_080_3 <= Relu_080;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_081_0 <= Relu_081;
		elsif(enb_b = '1') then
			reg_081_1 <= Relu_081;
		elsif(enb_c = '1') then
			reg_081_2 <= Relu_081;
		elsif(enb_d = '1') then
			reg_081_3 <= Relu_081;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_082_0 <= Relu_082;
		elsif(enb_b = '1') then
			reg_082_1 <= Relu_082;
		elsif(enb_c = '1') then
			reg_082_2 <= Relu_082;
		elsif(enb_d = '1') then
			reg_082_3 <= Relu_082;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_083_0 <= Relu_083;
		elsif(enb_b = '1') then
			reg_083_1 <= Relu_083;
		elsif(enb_c = '1') then
			reg_083_2 <= Relu_083;
		elsif(enb_d = '1') then
			reg_083_3 <= Relu_083;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_084_0 <= Relu_084;
		elsif(enb_b = '1') then
			reg_084_1 <= Relu_084;
		elsif(enb_c = '1') then
			reg_084_2 <= Relu_084;
		elsif(enb_d = '1') then
			reg_084_3 <= Relu_084;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_085_0 <= Relu_085;
		elsif(enb_b = '1') then
			reg_085_1 <= Relu_085;
		elsif(enb_c = '1') then
			reg_085_2 <= Relu_085;
		elsif(enb_d = '1') then
			reg_085_3 <= Relu_085;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_086_0 <= Relu_086;
		elsif(enb_b = '1') then
			reg_086_1 <= Relu_086;
		elsif(enb_c = '1') then
			reg_086_2 <= Relu_086;
		elsif(enb_d = '1') then
			reg_086_3 <= Relu_086;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_087_0 <= Relu_087;
		elsif(enb_b = '1') then
			reg_087_1 <= Relu_087;
		elsif(enb_c = '1') then
			reg_087_2 <= Relu_087;
		elsif(enb_d = '1') then
			reg_087_3 <= Relu_087;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_088_0 <= Relu_088;
		elsif(enb_b = '1') then
			reg_088_1 <= Relu_088;
		elsif(enb_c = '1') then
			reg_088_2 <= Relu_088;
		elsif(enb_d = '1') then
			reg_088_3 <= Relu_088;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_089_0 <= Relu_089;
		elsif(enb_b = '1') then
			reg_089_1 <= Relu_089;
		elsif(enb_c = '1') then
			reg_089_2 <= Relu_089;
		elsif(enb_d = '1') then
			reg_089_3 <= Relu_089;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_090_0 <= Relu_090;
		elsif(enb_b = '1') then
			reg_090_1 <= Relu_090;
		elsif(enb_c = '1') then
			reg_090_2 <= Relu_090;
		elsif(enb_d = '1') then
			reg_090_3 <= Relu_090;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_091_0 <= Relu_091;
		elsif(enb_b = '1') then
			reg_091_1 <= Relu_091;
		elsif(enb_c = '1') then
			reg_091_2 <= Relu_091;
		elsif(enb_d = '1') then
			reg_091_3 <= Relu_091;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_092_0 <= Relu_092;
		elsif(enb_b = '1') then
			reg_092_1 <= Relu_092;
		elsif(enb_c = '1') then
			reg_092_2 <= Relu_092;
		elsif(enb_d = '1') then
			reg_092_3 <= Relu_092;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_093_0 <= Relu_093;
		elsif(enb_b = '1') then
			reg_093_1 <= Relu_093;
		elsif(enb_c = '1') then
			reg_093_2 <= Relu_093;
		elsif(enb_d = '1') then
			reg_093_3 <= Relu_093;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_094_0 <= Relu_094;
		elsif(enb_b = '1') then
			reg_094_1 <= Relu_094;
		elsif(enb_c = '1') then
			reg_094_2 <= Relu_094;
		elsif(enb_d = '1') then
			reg_094_3 <= Relu_094;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_095_0 <= Relu_095;
		elsif(enb_b = '1') then
			reg_095_1 <= Relu_095;
		elsif(enb_c = '1') then
			reg_095_2 <= Relu_095;
		elsif(enb_d = '1') then
			reg_095_3 <= Relu_095;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_096_0 <= Relu_096;
		elsif(enb_b = '1') then
			reg_096_1 <= Relu_096;
		elsif(enb_c = '1') then
			reg_096_2 <= Relu_096;
		elsif(enb_d = '1') then
			reg_096_3 <= Relu_096;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_097_0 <= Relu_097;
		elsif(enb_b = '1') then
			reg_097_1 <= Relu_097;
		elsif(enb_c = '1') then
			reg_097_2 <= Relu_097;
		elsif(enb_d = '1') then
			reg_097_3 <= Relu_097;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_098_0 <= Relu_098;
		elsif(enb_b = '1') then
			reg_098_1 <= Relu_098;
		elsif(enb_c = '1') then
			reg_098_2 <= Relu_098;
		elsif(enb_d = '1') then
			reg_098_3 <= Relu_098;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_099_0 <= Relu_099;
		elsif(enb_b = '1') then
			reg_099_1 <= Relu_099;
		elsif(enb_c = '1') then
			reg_099_2 <= Relu_099;
		elsif(enb_d = '1') then
			reg_099_3 <= Relu_099;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_100_0 <= Relu_100;
		elsif(enb_b = '1') then
			reg_100_1 <= Relu_100;
		elsif(enb_c = '1') then
			reg_100_2 <= Relu_100;
		elsif(enb_d = '1') then
			reg_100_3 <= Relu_100;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_101_0 <= Relu_101;
		elsif(enb_b = '1') then
			reg_101_1 <= Relu_101;
		elsif(enb_c = '1') then
			reg_101_2 <= Relu_101;
		elsif(enb_d = '1') then
			reg_101_3 <= Relu_101;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_102_0 <= Relu_102;
		elsif(enb_b = '1') then
			reg_102_1 <= Relu_102;
		elsif(enb_c = '1') then
			reg_102_2 <= Relu_102;
		elsif(enb_d = '1') then
			reg_102_3 <= Relu_102;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_103_0 <= Relu_103;
		elsif(enb_b = '1') then
			reg_103_1 <= Relu_103;
		elsif(enb_c = '1') then
			reg_103_2 <= Relu_103;
		elsif(enb_d = '1') then
			reg_103_3 <= Relu_103;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_104_0 <= Relu_104;
		elsif(enb_b = '1') then
			reg_104_1 <= Relu_104;
		elsif(enb_c = '1') then
			reg_104_2 <= Relu_104;
		elsif(enb_d = '1') then
			reg_104_3 <= Relu_104;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_105_0 <= Relu_105;
		elsif(enb_b = '1') then
			reg_105_1 <= Relu_105;
		elsif(enb_c = '1') then
			reg_105_2 <= Relu_105;
		elsif(enb_d = '1') then
			reg_105_3 <= Relu_105;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_106_0 <= Relu_106;
		elsif(enb_b = '1') then
			reg_106_1 <= Relu_106;
		elsif(enb_c = '1') then
			reg_106_2 <= Relu_106;
		elsif(enb_d = '1') then
			reg_106_3 <= Relu_106;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_107_0 <= Relu_107;
		elsif(enb_b = '1') then
			reg_107_1 <= Relu_107;
		elsif(enb_c = '1') then
			reg_107_2 <= Relu_107;
		elsif(enb_d = '1') then
			reg_107_3 <= Relu_107;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_108_0 <= Relu_108;
		elsif(enb_b = '1') then
			reg_108_1 <= Relu_108;
		elsif(enb_c = '1') then
			reg_108_2 <= Relu_108;
		elsif(enb_d = '1') then
			reg_108_3 <= Relu_108;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_109_0 <= Relu_109;
		elsif(enb_b = '1') then
			reg_109_1 <= Relu_109;
		elsif(enb_c = '1') then
			reg_109_2 <= Relu_109;
		elsif(enb_d = '1') then
			reg_109_3 <= Relu_109;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_110_0 <= Relu_110;
		elsif(enb_b = '1') then
			reg_110_1 <= Relu_110;
		elsif(enb_c = '1') then
			reg_110_2 <= Relu_110;
		elsif(enb_d = '1') then
			reg_110_3 <= Relu_110;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_111_0 <= Relu_111;
		elsif(enb_b = '1') then
			reg_111_1 <= Relu_111;
		elsif(enb_c = '1') then
			reg_111_2 <= Relu_111;
		elsif(enb_d = '1') then
			reg_111_3 <= Relu_111;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_112_0 <= Relu_112;
		elsif(enb_b = '1') then
			reg_112_1 <= Relu_112;
		elsif(enb_c = '1') then
			reg_112_2 <= Relu_112;
		elsif(enb_d = '1') then
			reg_112_3 <= Relu_112;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_113_0 <= Relu_113;
		elsif(enb_b = '1') then
			reg_113_1 <= Relu_113;
		elsif(enb_c = '1') then
			reg_113_2 <= Relu_113;
		elsif(enb_d = '1') then
			reg_113_3 <= Relu_113;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_114_0 <= Relu_114;
		elsif(enb_b = '1') then
			reg_114_1 <= Relu_114;
		elsif(enb_c = '1') then
			reg_114_2 <= Relu_114;
		elsif(enb_d = '1') then
			reg_114_3 <= Relu_114;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_115_0 <= Relu_115;
		elsif(enb_b = '1') then
			reg_115_1 <= Relu_115;
		elsif(enb_c = '1') then
			reg_115_2 <= Relu_115;
		elsif(enb_d = '1') then
			reg_115_3 <= Relu_115;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_116_0 <= Relu_116;
		elsif(enb_b = '1') then
			reg_116_1 <= Relu_116;
		elsif(enb_c = '1') then
			reg_116_2 <= Relu_116;
		elsif(enb_d = '1') then
			reg_116_3 <= Relu_116;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_117_0 <= Relu_117;
		elsif(enb_b = '1') then
			reg_117_1 <= Relu_117;
		elsif(enb_c = '1') then
			reg_117_2 <= Relu_117;
		elsif(enb_d = '1') then
			reg_117_3 <= Relu_117;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_118_0 <= Relu_118;
		elsif(enb_b = '1') then
			reg_118_1 <= Relu_118;
		elsif(enb_c = '1') then
			reg_118_2 <= Relu_118;
		elsif(enb_d = '1') then
			reg_118_3 <= Relu_118;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_119_0 <= Relu_119;
		elsif(enb_b = '1') then
			reg_119_1 <= Relu_119;
		elsif(enb_c = '1') then
			reg_119_2 <= Relu_119;
		elsif(enb_d = '1') then
			reg_119_3 <= Relu_119;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_120_0 <= Relu_120;
		elsif(enb_b = '1') then
			reg_120_1 <= Relu_120;
		elsif(enb_c = '1') then
			reg_120_2 <= Relu_120;
		elsif(enb_d = '1') then
			reg_120_3 <= Relu_120;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_121_0 <= Relu_121;
		elsif(enb_b = '1') then
			reg_121_1 <= Relu_121;
		elsif(enb_c = '1') then
			reg_121_2 <= Relu_121;
		elsif(enb_d = '1') then
			reg_121_3 <= Relu_121;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_122_0 <= Relu_122;
		elsif(enb_b = '1') then
			reg_122_1 <= Relu_122;
		elsif(enb_c = '1') then
			reg_122_2 <= Relu_122;
		elsif(enb_d = '1') then
			reg_122_3 <= Relu_122;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_123_0 <= Relu_123;
		elsif(enb_b = '1') then
			reg_123_1 <= Relu_123;
		elsif(enb_c = '1') then
			reg_123_2 <= Relu_123;
		elsif(enb_d = '1') then
			reg_123_3 <= Relu_123;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_124_0 <= Relu_124;
		elsif(enb_b = '1') then
			reg_124_1 <= Relu_124;
		elsif(enb_c = '1') then
			reg_124_2 <= Relu_124;
		elsif(enb_d = '1') then
			reg_124_3 <= Relu_124;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_125_0 <= Relu_125;
		elsif(enb_b = '1') then
			reg_125_1 <= Relu_125;
		elsif(enb_c = '1') then
			reg_125_2 <= Relu_125;
		elsif(enb_d = '1') then
			reg_125_3 <= Relu_125;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_126_0 <= Relu_126;
		elsif(enb_b = '1') then
			reg_126_1 <= Relu_126;
		elsif(enb_c = '1') then
			reg_126_2 <= Relu_126;
		elsif(enb_d = '1') then
			reg_126_3 <= Relu_126;
		end if;
	end if;  
end process;

process( RST_N, CLK )  
begin 
	if( CLK'event and CLK = '1' ) then
		if(enb_a = '1') then
			reg_127_0 <= Relu_127;
		elsif(enb_b = '1') then
			reg_127_1 <= Relu_127;
		elsif(enb_c = '1') then
			reg_127_2 <= Relu_127;
		elsif(enb_d = '1') then
			reg_127_3 <= Relu_127;
		end if;
	end if;  
end process;


-------------------------------------------------------
ROM_W9_inst : ROM_W9 port map (
  CLK => CLK,
  ADDR => W9Addr,
  DATA => W9Data );

---------------------------------------------------------
coeff_000 <= W9Data( 1535 downto 1533 ) when W9Sel = "00" else W9Data( 1532 downto 1530 ) when W9Sel = "01" else W9Data( 1529 downto 1527 ) when W9Sel = "10" else W9Data( 1526 downto 1524 );

coeff_001 <= W9Data( 1523 downto 1521 ) when W9Sel = "00" else W9Data( 1520 downto 1518 ) when W9Sel = "01" else W9Data( 1517 downto 1515 ) when W9Sel = "10" else W9Data( 1514 downto 1512 );

coeff_002 <= W9Data( 1511 downto 1509 ) when W9Sel = "00" else W9Data( 1508 downto 1506 ) when W9Sel = "01" else W9Data( 1505 downto 1503 ) when W9Sel = "10" else W9Data( 1502 downto 1500 );

coeff_003 <= W9Data( 1499 downto 1497 ) when W9Sel = "00" else W9Data( 1496 downto 1494 ) when W9Sel = "01" else W9Data( 1493 downto 1491 ) when W9Sel = "10" else W9Data( 1490 downto 1488 );

coeff_004 <= W9Data( 1487 downto 1485 ) when W9Sel = "00" else W9Data( 1484 downto 1482 ) when W9Sel = "01" else W9Data( 1481 downto 1479 ) when W9Sel = "10" else W9Data( 1478 downto 1476 );

coeff_005 <= W9Data( 1475 downto 1473 ) when W9Sel = "00" else W9Data( 1472 downto 1470 ) when W9Sel = "01" else W9Data( 1469 downto 1467 ) when W9Sel = "10" else W9Data( 1466 downto 1464 );

coeff_006 <= W9Data( 1463 downto 1461 ) when W9Sel = "00" else W9Data( 1460 downto 1458 ) when W9Sel = "01" else W9Data( 1457 downto 1455 ) when W9Sel = "10" else W9Data( 1454 downto 1452 );

coeff_007 <= W9Data( 1451 downto 1449 ) when W9Sel = "00" else W9Data( 1448 downto 1446 ) when W9Sel = "01" else W9Data( 1445 downto 1443 ) when W9Sel = "10" else W9Data( 1442 downto 1440 );

coeff_008 <= W9Data( 1439 downto 1437 ) when W9Sel = "00" else W9Data( 1436 downto 1434 ) when W9Sel = "01" else W9Data( 1433 downto 1431 ) when W9Sel = "10" else W9Data( 1430 downto 1428 );

coeff_009 <= W9Data( 1427 downto 1425 ) when W9Sel = "00" else W9Data( 1424 downto 1422 ) when W9Sel = "01" else W9Data( 1421 downto 1419 ) when W9Sel = "10" else W9Data( 1418 downto 1416 );

coeff_010 <= W9Data( 1415 downto 1413 ) when W9Sel = "00" else W9Data( 1412 downto 1410 ) when W9Sel = "01" else W9Data( 1409 downto 1407 ) when W9Sel = "10" else W9Data( 1406 downto 1404 );

coeff_011 <= W9Data( 1403 downto 1401 ) when W9Sel = "00" else W9Data( 1400 downto 1398 ) when W9Sel = "01" else W9Data( 1397 downto 1395 ) when W9Sel = "10" else W9Data( 1394 downto 1392 );

coeff_012 <= W9Data( 1391 downto 1389 ) when W9Sel = "00" else W9Data( 1388 downto 1386 ) when W9Sel = "01" else W9Data( 1385 downto 1383 ) when W9Sel = "10" else W9Data( 1382 downto 1380 );

coeff_013 <= W9Data( 1379 downto 1377 ) when W9Sel = "00" else W9Data( 1376 downto 1374 ) when W9Sel = "01" else W9Data( 1373 downto 1371 ) when W9Sel = "10" else W9Data( 1370 downto 1368 );

coeff_014 <= W9Data( 1367 downto 1365 ) when W9Sel = "00" else W9Data( 1364 downto 1362 ) when W9Sel = "01" else W9Data( 1361 downto 1359 ) when W9Sel = "10" else W9Data( 1358 downto 1356 );

coeff_015 <= W9Data( 1355 downto 1353 ) when W9Sel = "00" else W9Data( 1352 downto 1350 ) when W9Sel = "01" else W9Data( 1349 downto 1347 ) when W9Sel = "10" else W9Data( 1346 downto 1344 );

coeff_016 <= W9Data( 1343 downto 1341 ) when W9Sel = "00" else W9Data( 1340 downto 1338 ) when W9Sel = "01" else W9Data( 1337 downto 1335 ) when W9Sel = "10" else W9Data( 1334 downto 1332 );

coeff_017 <= W9Data( 1331 downto 1329 ) when W9Sel = "00" else W9Data( 1328 downto 1326 ) when W9Sel = "01" else W9Data( 1325 downto 1323 ) when W9Sel = "10" else W9Data( 1322 downto 1320 );

coeff_018 <= W9Data( 1319 downto 1317 ) when W9Sel = "00" else W9Data( 1316 downto 1314 ) when W9Sel = "01" else W9Data( 1313 downto 1311 ) when W9Sel = "10" else W9Data( 1310 downto 1308 );

coeff_019 <= W9Data( 1307 downto 1305 ) when W9Sel = "00" else W9Data( 1304 downto 1302 ) when W9Sel = "01" else W9Data( 1301 downto 1299 ) when W9Sel = "10" else W9Data( 1298 downto 1296 );

coeff_020 <= W9Data( 1295 downto 1293 ) when W9Sel = "00" else W9Data( 1292 downto 1290 ) when W9Sel = "01" else W9Data( 1289 downto 1287 ) when W9Sel = "10" else W9Data( 1286 downto 1284 );

coeff_021 <= W9Data( 1283 downto 1281 ) when W9Sel = "00" else W9Data( 1280 downto 1278 ) when W9Sel = "01" else W9Data( 1277 downto 1275 ) when W9Sel = "10" else W9Data( 1274 downto 1272 );

coeff_022 <= W9Data( 1271 downto 1269 ) when W9Sel = "00" else W9Data( 1268 downto 1266 ) when W9Sel = "01" else W9Data( 1265 downto 1263 ) when W9Sel = "10" else W9Data( 1262 downto 1260 );

coeff_023 <= W9Data( 1259 downto 1257 ) when W9Sel = "00" else W9Data( 1256 downto 1254 ) when W9Sel = "01" else W9Data( 1253 downto 1251 ) when W9Sel = "10" else W9Data( 1250 downto 1248 );

coeff_024 <= W9Data( 1247 downto 1245 ) when W9Sel = "00" else W9Data( 1244 downto 1242 ) when W9Sel = "01" else W9Data( 1241 downto 1239 ) when W9Sel = "10" else W9Data( 1238 downto 1236 );

coeff_025 <= W9Data( 1235 downto 1233 ) when W9Sel = "00" else W9Data( 1232 downto 1230 ) when W9Sel = "01" else W9Data( 1229 downto 1227 ) when W9Sel = "10" else W9Data( 1226 downto 1224 );

coeff_026 <= W9Data( 1223 downto 1221 ) when W9Sel = "00" else W9Data( 1220 downto 1218 ) when W9Sel = "01" else W9Data( 1217 downto 1215 ) when W9Sel = "10" else W9Data( 1214 downto 1212 );

coeff_027 <= W9Data( 1211 downto 1209 ) when W9Sel = "00" else W9Data( 1208 downto 1206 ) when W9Sel = "01" else W9Data( 1205 downto 1203 ) when W9Sel = "10" else W9Data( 1202 downto 1200 );

coeff_028 <= W9Data( 1199 downto 1197 ) when W9Sel = "00" else W9Data( 1196 downto 1194 ) when W9Sel = "01" else W9Data( 1193 downto 1191 ) when W9Sel = "10" else W9Data( 1190 downto 1188 );

coeff_029 <= W9Data( 1187 downto 1185 ) when W9Sel = "00" else W9Data( 1184 downto 1182 ) when W9Sel = "01" else W9Data( 1181 downto 1179 ) when W9Sel = "10" else W9Data( 1178 downto 1176 );

coeff_030 <= W9Data( 1175 downto 1173 ) when W9Sel = "00" else W9Data( 1172 downto 1170 ) when W9Sel = "01" else W9Data( 1169 downto 1167 ) when W9Sel = "10" else W9Data( 1166 downto 1164 );

coeff_031 <= W9Data( 1163 downto 1161 ) when W9Sel = "00" else W9Data( 1160 downto 1158 ) when W9Sel = "01" else W9Data( 1157 downto 1155 ) when W9Sel = "10" else W9Data( 1154 downto 1152 );

coeff_032 <= W9Data( 1151 downto 1149 ) when W9Sel = "00" else W9Data( 1148 downto 1146 ) when W9Sel = "01" else W9Data( 1145 downto 1143 ) when W9Sel = "10" else W9Data( 1142 downto 1140 );

coeff_033 <= W9Data( 1139 downto 1137 ) when W9Sel = "00" else W9Data( 1136 downto 1134 ) when W9Sel = "01" else W9Data( 1133 downto 1131 ) when W9Sel = "10" else W9Data( 1130 downto 1128 );

coeff_034 <= W9Data( 1127 downto 1125 ) when W9Sel = "00" else W9Data( 1124 downto 1122 ) when W9Sel = "01" else W9Data( 1121 downto 1119 ) when W9Sel = "10" else W9Data( 1118 downto 1116 );

coeff_035 <= W9Data( 1115 downto 1113 ) when W9Sel = "00" else W9Data( 1112 downto 1110 ) when W9Sel = "01" else W9Data( 1109 downto 1107 ) when W9Sel = "10" else W9Data( 1106 downto 1104 );

coeff_036 <= W9Data( 1103 downto 1101 ) when W9Sel = "00" else W9Data( 1100 downto 1098 ) when W9Sel = "01" else W9Data( 1097 downto 1095 ) when W9Sel = "10" else W9Data( 1094 downto 1092 );

coeff_037 <= W9Data( 1091 downto 1089 ) when W9Sel = "00" else W9Data( 1088 downto 1086 ) when W9Sel = "01" else W9Data( 1085 downto 1083 ) when W9Sel = "10" else W9Data( 1082 downto 1080 );

coeff_038 <= W9Data( 1079 downto 1077 ) when W9Sel = "00" else W9Data( 1076 downto 1074 ) when W9Sel = "01" else W9Data( 1073 downto 1071 ) when W9Sel = "10" else W9Data( 1070 downto 1068 );

coeff_039 <= W9Data( 1067 downto 1065 ) when W9Sel = "00" else W9Data( 1064 downto 1062 ) when W9Sel = "01" else W9Data( 1061 downto 1059 ) when W9Sel = "10" else W9Data( 1058 downto 1056 );

coeff_040 <= W9Data( 1055 downto 1053 ) when W9Sel = "00" else W9Data( 1052 downto 1050 ) when W9Sel = "01" else W9Data( 1049 downto 1047 ) when W9Sel = "10" else W9Data( 1046 downto 1044 );

coeff_041 <= W9Data( 1043 downto 1041 ) when W9Sel = "00" else W9Data( 1040 downto 1038 ) when W9Sel = "01" else W9Data( 1037 downto 1035 ) when W9Sel = "10" else W9Data( 1034 downto 1032 );

coeff_042 <= W9Data( 1031 downto 1029 ) when W9Sel = "00" else W9Data( 1028 downto 1026 ) when W9Sel = "01" else W9Data( 1025 downto 1023 ) when W9Sel = "10" else W9Data( 1022 downto 1020 );

coeff_043 <= W9Data( 1019 downto 1017 ) when W9Sel = "00" else W9Data( 1016 downto 1014 ) when W9Sel = "01" else W9Data( 1013 downto 1011 ) when W9Sel = "10" else W9Data( 1010 downto 1008 );

coeff_044 <= W9Data( 1007 downto 1005 ) when W9Sel = "00" else W9Data( 1004 downto 1002 ) when W9Sel = "01" else W9Data( 1001 downto 999 ) when W9Sel = "10" else W9Data( 998 downto 996 );

coeff_045 <= W9Data( 995 downto 993 ) when W9Sel = "00" else W9Data( 992 downto 990 ) when W9Sel = "01" else W9Data( 989 downto 987 ) when W9Sel = "10" else W9Data( 986 downto 984 );

coeff_046 <= W9Data( 983 downto 981 ) when W9Sel = "00" else W9Data( 980 downto 978 ) when W9Sel = "01" else W9Data( 977 downto 975 ) when W9Sel = "10" else W9Data( 974 downto 972 );

coeff_047 <= W9Data( 971 downto 969 ) when W9Sel = "00" else W9Data( 968 downto 966 ) when W9Sel = "01" else W9Data( 965 downto 963 ) when W9Sel = "10" else W9Data( 962 downto 960 );

coeff_048 <= W9Data( 959 downto 957 ) when W9Sel = "00" else W9Data( 956 downto 954 ) when W9Sel = "01" else W9Data( 953 downto 951 ) when W9Sel = "10" else W9Data( 950 downto 948 );

coeff_049 <= W9Data( 947 downto 945 ) when W9Sel = "00" else W9Data( 944 downto 942 ) when W9Sel = "01" else W9Data( 941 downto 939 ) when W9Sel = "10" else W9Data( 938 downto 936 );

coeff_050 <= W9Data( 935 downto 933 ) when W9Sel = "00" else W9Data( 932 downto 930 ) when W9Sel = "01" else W9Data( 929 downto 927 ) when W9Sel = "10" else W9Data( 926 downto 924 );

coeff_051 <= W9Data( 923 downto 921 ) when W9Sel = "00" else W9Data( 920 downto 918 ) when W9Sel = "01" else W9Data( 917 downto 915 ) when W9Sel = "10" else W9Data( 914 downto 912 );

coeff_052 <= W9Data( 911 downto 909 ) when W9Sel = "00" else W9Data( 908 downto 906 ) when W9Sel = "01" else W9Data( 905 downto 903 ) when W9Sel = "10" else W9Data( 902 downto 900 );

coeff_053 <= W9Data( 899 downto 897 ) when W9Sel = "00" else W9Data( 896 downto 894 ) when W9Sel = "01" else W9Data( 893 downto 891 ) when W9Sel = "10" else W9Data( 890 downto 888 );

coeff_054 <= W9Data( 887 downto 885 ) when W9Sel = "00" else W9Data( 884 downto 882 ) when W9Sel = "01" else W9Data( 881 downto 879 ) when W9Sel = "10" else W9Data( 878 downto 876 );

coeff_055 <= W9Data( 875 downto 873 ) when W9Sel = "00" else W9Data( 872 downto 870 ) when W9Sel = "01" else W9Data( 869 downto 867 ) when W9Sel = "10" else W9Data( 866 downto 864 );

coeff_056 <= W9Data( 863 downto 861 ) when W9Sel = "00" else W9Data( 860 downto 858 ) when W9Sel = "01" else W9Data( 857 downto 855 ) when W9Sel = "10" else W9Data( 854 downto 852 );

coeff_057 <= W9Data( 851 downto 849 ) when W9Sel = "00" else W9Data( 848 downto 846 ) when W9Sel = "01" else W9Data( 845 downto 843 ) when W9Sel = "10" else W9Data( 842 downto 840 );

coeff_058 <= W9Data( 839 downto 837 ) when W9Sel = "00" else W9Data( 836 downto 834 ) when W9Sel = "01" else W9Data( 833 downto 831 ) when W9Sel = "10" else W9Data( 830 downto 828 );

coeff_059 <= W9Data( 827 downto 825 ) when W9Sel = "00" else W9Data( 824 downto 822 ) when W9Sel = "01" else W9Data( 821 downto 819 ) when W9Sel = "10" else W9Data( 818 downto 816 );

coeff_060 <= W9Data( 815 downto 813 ) when W9Sel = "00" else W9Data( 812 downto 810 ) when W9Sel = "01" else W9Data( 809 downto 807 ) when W9Sel = "10" else W9Data( 806 downto 804 );

coeff_061 <= W9Data( 803 downto 801 ) when W9Sel = "00" else W9Data( 800 downto 798 ) when W9Sel = "01" else W9Data( 797 downto 795 ) when W9Sel = "10" else W9Data( 794 downto 792 );

coeff_062 <= W9Data( 791 downto 789 ) when W9Sel = "00" else W9Data( 788 downto 786 ) when W9Sel = "01" else W9Data( 785 downto 783 ) when W9Sel = "10" else W9Data( 782 downto 780 );

coeff_063 <= W9Data( 779 downto 777 ) when W9Sel = "00" else W9Data( 776 downto 774 ) when W9Sel = "01" else W9Data( 773 downto 771 ) when W9Sel = "10" else W9Data( 770 downto 768 );

coeff_064 <= W9Data( 767 downto 765 ) when W9Sel = "00" else W9Data( 764 downto 762 ) when W9Sel = "01" else W9Data( 761 downto 759 ) when W9Sel = "10" else W9Data( 758 downto 756 );

coeff_065 <= W9Data( 755 downto 753 ) when W9Sel = "00" else W9Data( 752 downto 750 ) when W9Sel = "01" else W9Data( 749 downto 747 ) when W9Sel = "10" else W9Data( 746 downto 744 );

coeff_066 <= W9Data( 743 downto 741 ) when W9Sel = "00" else W9Data( 740 downto 738 ) when W9Sel = "01" else W9Data( 737 downto 735 ) when W9Sel = "10" else W9Data( 734 downto 732 );

coeff_067 <= W9Data( 731 downto 729 ) when W9Sel = "00" else W9Data( 728 downto 726 ) when W9Sel = "01" else W9Data( 725 downto 723 ) when W9Sel = "10" else W9Data( 722 downto 720 );

coeff_068 <= W9Data( 719 downto 717 ) when W9Sel = "00" else W9Data( 716 downto 714 ) when W9Sel = "01" else W9Data( 713 downto 711 ) when W9Sel = "10" else W9Data( 710 downto 708 );

coeff_069 <= W9Data( 707 downto 705 ) when W9Sel = "00" else W9Data( 704 downto 702 ) when W9Sel = "01" else W9Data( 701 downto 699 ) when W9Sel = "10" else W9Data( 698 downto 696 );

coeff_070 <= W9Data( 695 downto 693 ) when W9Sel = "00" else W9Data( 692 downto 690 ) when W9Sel = "01" else W9Data( 689 downto 687 ) when W9Sel = "10" else W9Data( 686 downto 684 );

coeff_071 <= W9Data( 683 downto 681 ) when W9Sel = "00" else W9Data( 680 downto 678 ) when W9Sel = "01" else W9Data( 677 downto 675 ) when W9Sel = "10" else W9Data( 674 downto 672 );

coeff_072 <= W9Data( 671 downto 669 ) when W9Sel = "00" else W9Data( 668 downto 666 ) when W9Sel = "01" else W9Data( 665 downto 663 ) when W9Sel = "10" else W9Data( 662 downto 660 );

coeff_073 <= W9Data( 659 downto 657 ) when W9Sel = "00" else W9Data( 656 downto 654 ) when W9Sel = "01" else W9Data( 653 downto 651 ) when W9Sel = "10" else W9Data( 650 downto 648 );

coeff_074 <= W9Data( 647 downto 645 ) when W9Sel = "00" else W9Data( 644 downto 642 ) when W9Sel = "01" else W9Data( 641 downto 639 ) when W9Sel = "10" else W9Data( 638 downto 636 );

coeff_075 <= W9Data( 635 downto 633 ) when W9Sel = "00" else W9Data( 632 downto 630 ) when W9Sel = "01" else W9Data( 629 downto 627 ) when W9Sel = "10" else W9Data( 626 downto 624 );

coeff_076 <= W9Data( 623 downto 621 ) when W9Sel = "00" else W9Data( 620 downto 618 ) when W9Sel = "01" else W9Data( 617 downto 615 ) when W9Sel = "10" else W9Data( 614 downto 612 );

coeff_077 <= W9Data( 611 downto 609 ) when W9Sel = "00" else W9Data( 608 downto 606 ) when W9Sel = "01" else W9Data( 605 downto 603 ) when W9Sel = "10" else W9Data( 602 downto 600 );

coeff_078 <= W9Data( 599 downto 597 ) when W9Sel = "00" else W9Data( 596 downto 594 ) when W9Sel = "01" else W9Data( 593 downto 591 ) when W9Sel = "10" else W9Data( 590 downto 588 );

coeff_079 <= W9Data( 587 downto 585 ) when W9Sel = "00" else W9Data( 584 downto 582 ) when W9Sel = "01" else W9Data( 581 downto 579 ) when W9Sel = "10" else W9Data( 578 downto 576 );

coeff_080 <= W9Data( 575 downto 573 ) when W9Sel = "00" else W9Data( 572 downto 570 ) when W9Sel = "01" else W9Data( 569 downto 567 ) when W9Sel = "10" else W9Data( 566 downto 564 );

coeff_081 <= W9Data( 563 downto 561 ) when W9Sel = "00" else W9Data( 560 downto 558 ) when W9Sel = "01" else W9Data( 557 downto 555 ) when W9Sel = "10" else W9Data( 554 downto 552 );

coeff_082 <= W9Data( 551 downto 549 ) when W9Sel = "00" else W9Data( 548 downto 546 ) when W9Sel = "01" else W9Data( 545 downto 543 ) when W9Sel = "10" else W9Data( 542 downto 540 );

coeff_083 <= W9Data( 539 downto 537 ) when W9Sel = "00" else W9Data( 536 downto 534 ) when W9Sel = "01" else W9Data( 533 downto 531 ) when W9Sel = "10" else W9Data( 530 downto 528 );

coeff_084 <= W9Data( 527 downto 525 ) when W9Sel = "00" else W9Data( 524 downto 522 ) when W9Sel = "01" else W9Data( 521 downto 519 ) when W9Sel = "10" else W9Data( 518 downto 516 );

coeff_085 <= W9Data( 515 downto 513 ) when W9Sel = "00" else W9Data( 512 downto 510 ) when W9Sel = "01" else W9Data( 509 downto 507 ) when W9Sel = "10" else W9Data( 506 downto 504 );

coeff_086 <= W9Data( 503 downto 501 ) when W9Sel = "00" else W9Data( 500 downto 498 ) when W9Sel = "01" else W9Data( 497 downto 495 ) when W9Sel = "10" else W9Data( 494 downto 492 );

coeff_087 <= W9Data( 491 downto 489 ) when W9Sel = "00" else W9Data( 488 downto 486 ) when W9Sel = "01" else W9Data( 485 downto 483 ) when W9Sel = "10" else W9Data( 482 downto 480 );

coeff_088 <= W9Data( 479 downto 477 ) when W9Sel = "00" else W9Data( 476 downto 474 ) when W9Sel = "01" else W9Data( 473 downto 471 ) when W9Sel = "10" else W9Data( 470 downto 468 );

coeff_089 <= W9Data( 467 downto 465 ) when W9Sel = "00" else W9Data( 464 downto 462 ) when W9Sel = "01" else W9Data( 461 downto 459 ) when W9Sel = "10" else W9Data( 458 downto 456 );

coeff_090 <= W9Data( 455 downto 453 ) when W9Sel = "00" else W9Data( 452 downto 450 ) when W9Sel = "01" else W9Data( 449 downto 447 ) when W9Sel = "10" else W9Data( 446 downto 444 );

coeff_091 <= W9Data( 443 downto 441 ) when W9Sel = "00" else W9Data( 440 downto 438 ) when W9Sel = "01" else W9Data( 437 downto 435 ) when W9Sel = "10" else W9Data( 434 downto 432 );

coeff_092 <= W9Data( 431 downto 429 ) when W9Sel = "00" else W9Data( 428 downto 426 ) when W9Sel = "01" else W9Data( 425 downto 423 ) when W9Sel = "10" else W9Data( 422 downto 420 );

coeff_093 <= W9Data( 419 downto 417 ) when W9Sel = "00" else W9Data( 416 downto 414 ) when W9Sel = "01" else W9Data( 413 downto 411 ) when W9Sel = "10" else W9Data( 410 downto 408 );

coeff_094 <= W9Data( 407 downto 405 ) when W9Sel = "00" else W9Data( 404 downto 402 ) when W9Sel = "01" else W9Data( 401 downto 399 ) when W9Sel = "10" else W9Data( 398 downto 396 );

coeff_095 <= W9Data( 395 downto 393 ) when W9Sel = "00" else W9Data( 392 downto 390 ) when W9Sel = "01" else W9Data( 389 downto 387 ) when W9Sel = "10" else W9Data( 386 downto 384 );

coeff_096 <= W9Data( 383 downto 381 ) when W9Sel = "00" else W9Data( 380 downto 378 ) when W9Sel = "01" else W9Data( 377 downto 375 ) when W9Sel = "10" else W9Data( 374 downto 372 );

coeff_097 <= W9Data( 371 downto 369 ) when W9Sel = "00" else W9Data( 368 downto 366 ) when W9Sel = "01" else W9Data( 365 downto 363 ) when W9Sel = "10" else W9Data( 362 downto 360 );

coeff_098 <= W9Data( 359 downto 357 ) when W9Sel = "00" else W9Data( 356 downto 354 ) when W9Sel = "01" else W9Data( 353 downto 351 ) when W9Sel = "10" else W9Data( 350 downto 348 );

coeff_099 <= W9Data( 347 downto 345 ) when W9Sel = "00" else W9Data( 344 downto 342 ) when W9Sel = "01" else W9Data( 341 downto 339 ) when W9Sel = "10" else W9Data( 338 downto 336 );

coeff_100 <= W9Data( 335 downto 333 ) when W9Sel = "00" else W9Data( 332 downto 330 ) when W9Sel = "01" else W9Data( 329 downto 327 ) when W9Sel = "10" else W9Data( 326 downto 324 );

coeff_101 <= W9Data( 323 downto 321 ) when W9Sel = "00" else W9Data( 320 downto 318 ) when W9Sel = "01" else W9Data( 317 downto 315 ) when W9Sel = "10" else W9Data( 314 downto 312 );

coeff_102 <= W9Data( 311 downto 309 ) when W9Sel = "00" else W9Data( 308 downto 306 ) when W9Sel = "01" else W9Data( 305 downto 303 ) when W9Sel = "10" else W9Data( 302 downto 300 );

coeff_103 <= W9Data( 299 downto 297 ) when W9Sel = "00" else W9Data( 296 downto 294 ) when W9Sel = "01" else W9Data( 293 downto 291 ) when W9Sel = "10" else W9Data( 290 downto 288 );

coeff_104 <= W9Data( 287 downto 285 ) when W9Sel = "00" else W9Data( 284 downto 282 ) when W9Sel = "01" else W9Data( 281 downto 279 ) when W9Sel = "10" else W9Data( 278 downto 276 );

coeff_105 <= W9Data( 275 downto 273 ) when W9Sel = "00" else W9Data( 272 downto 270 ) when W9Sel = "01" else W9Data( 269 downto 267 ) when W9Sel = "10" else W9Data( 266 downto 264 );

coeff_106 <= W9Data( 263 downto 261 ) when W9Sel = "00" else W9Data( 260 downto 258 ) when W9Sel = "01" else W9Data( 257 downto 255 ) when W9Sel = "10" else W9Data( 254 downto 252 );

coeff_107 <= W9Data( 251 downto 249 ) when W9Sel = "00" else W9Data( 248 downto 246 ) when W9Sel = "01" else W9Data( 245 downto 243 ) when W9Sel = "10" else W9Data( 242 downto 240 );

coeff_108 <= W9Data( 239 downto 237 ) when W9Sel = "00" else W9Data( 236 downto 234 ) when W9Sel = "01" else W9Data( 233 downto 231 ) when W9Sel = "10" else W9Data( 230 downto 228 );

coeff_109 <= W9Data( 227 downto 225 ) when W9Sel = "00" else W9Data( 224 downto 222 ) when W9Sel = "01" else W9Data( 221 downto 219 ) when W9Sel = "10" else W9Data( 218 downto 216 );

coeff_110 <= W9Data( 215 downto 213 ) when W9Sel = "00" else W9Data( 212 downto 210 ) when W9Sel = "01" else W9Data( 209 downto 207 ) when W9Sel = "10" else W9Data( 206 downto 204 );

coeff_111 <= W9Data( 203 downto 201 ) when W9Sel = "00" else W9Data( 200 downto 198 ) when W9Sel = "01" else W9Data( 197 downto 195 ) when W9Sel = "10" else W9Data( 194 downto 192 );

coeff_112 <= W9Data( 191 downto 189 ) when W9Sel = "00" else W9Data( 188 downto 186 ) when W9Sel = "01" else W9Data( 185 downto 183 ) when W9Sel = "10" else W9Data( 182 downto 180 );

coeff_113 <= W9Data( 179 downto 177 ) when W9Sel = "00" else W9Data( 176 downto 174 ) when W9Sel = "01" else W9Data( 173 downto 171 ) when W9Sel = "10" else W9Data( 170 downto 168 );

coeff_114 <= W9Data( 167 downto 165 ) when W9Sel = "00" else W9Data( 164 downto 162 ) when W9Sel = "01" else W9Data( 161 downto 159 ) when W9Sel = "10" else W9Data( 158 downto 156 );

coeff_115 <= W9Data( 155 downto 153 ) when W9Sel = "00" else W9Data( 152 downto 150 ) when W9Sel = "01" else W9Data( 149 downto 147 ) when W9Sel = "10" else W9Data( 146 downto 144 );

coeff_116 <= W9Data( 143 downto 141 ) when W9Sel = "00" else W9Data( 140 downto 138 ) when W9Sel = "01" else W9Data( 137 downto 135 ) when W9Sel = "10" else W9Data( 134 downto 132 );

coeff_117 <= W9Data( 131 downto 129 ) when W9Sel = "00" else W9Data( 128 downto 126 ) when W9Sel = "01" else W9Data( 125 downto 123 ) when W9Sel = "10" else W9Data( 122 downto 120 );

coeff_118 <= W9Data( 119 downto 117 ) when W9Sel = "00" else W9Data( 116 downto 114 ) when W9Sel = "01" else W9Data( 113 downto 111 ) when W9Sel = "10" else W9Data( 110 downto 108 );

coeff_119 <= W9Data( 107 downto 105 ) when W9Sel = "00" else W9Data( 104 downto 102 ) when W9Sel = "01" else W9Data( 101 downto 99 ) when W9Sel = "10" else W9Data( 98 downto 96 );

coeff_120 <= W9Data( 95 downto 93 ) when W9Sel = "00" else W9Data( 92 downto 90 ) when W9Sel = "01" else W9Data( 89 downto 87 ) when W9Sel = "10" else W9Data( 86 downto 84 );

coeff_121 <= W9Data( 83 downto 81 ) when W9Sel = "00" else W9Data( 80 downto 78 ) when W9Sel = "01" else W9Data( 77 downto 75 ) when W9Sel = "10" else W9Data( 74 downto 72 );

coeff_122 <= W9Data( 71 downto 69 ) when W9Sel = "00" else W9Data( 68 downto 66 ) when W9Sel = "01" else W9Data( 65 downto 63 ) when W9Sel = "10" else W9Data( 62 downto 60 );

coeff_123 <= W9Data( 59 downto 57 ) when W9Sel = "00" else W9Data( 56 downto 54 ) when W9Sel = "01" else W9Data( 53 downto 51 ) when W9Sel = "10" else W9Data( 50 downto 48 );

coeff_124 <= W9Data( 47 downto 45 ) when W9Sel = "00" else W9Data( 44 downto 42 ) when W9Sel = "01" else W9Data( 41 downto 39 ) when W9Sel = "10" else W9Data( 38 downto 36 );

coeff_125 <= W9Data( 35 downto 33 ) when W9Sel = "00" else W9Data( 32 downto 30 ) when W9Sel = "01" else W9Data( 29 downto 27 ) when W9Sel = "10" else W9Data( 26 downto 24 );

coeff_126 <= W9Data( 23 downto 21 ) when W9Sel = "00" else W9Data( 20 downto 18 ) when W9Sel = "01" else W9Data( 17 downto 15 ) when W9Sel = "10" else W9Data( 14 downto 12 );

coeff_127 <= W9Data( 11 downto 9 ) when W9Sel = "00" else W9Data( 8 downto 6 ) when W9Sel = "01" else W9Data( 5 downto 3 ) when W9Sel = "10" else W9Data( 2 downto 0 );


---------------------------------------------------------
dotpro_w9_000 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_000,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_000 );

dotpro_w9_001 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_001,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_001 );

dotpro_w9_002 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_002,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_002 );

dotpro_w9_003 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_003,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_003 );

dotpro_w9_004 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_004,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_004 );

dotpro_w9_005 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_005,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_005 );

dotpro_w9_006 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_006,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_006 );

dotpro_w9_007 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_007,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_007 );

dotpro_w9_008 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_008,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_008 );

dotpro_w9_009 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_009,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_009 );

dotpro_w9_010 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_010,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_010 );

dotpro_w9_011 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_011,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_011 );

dotpro_w9_012 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_012,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_012 );

dotpro_w9_013 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_013,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_013 );

dotpro_w9_014 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_014,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_014 );

dotpro_w9_015 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_015,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_015 );

dotpro_w9_016 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_016,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_016 );

dotpro_w9_017 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_017,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_017 );

dotpro_w9_018 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_018,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_018 );

dotpro_w9_019 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_019,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_019 );

dotpro_w9_020 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_020,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_020 );

dotpro_w9_021 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_021,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_021 );

dotpro_w9_022 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_022,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_022 );

dotpro_w9_023 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_023,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_023 );

dotpro_w9_024 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_024,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_024 );

dotpro_w9_025 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_025,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_025 );

dotpro_w9_026 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_026,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_026 );

dotpro_w9_027 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_027,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_027 );

dotpro_w9_028 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_028,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_028 );

dotpro_w9_029 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_029,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_029 );

dotpro_w9_030 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_030,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_030 );

dotpro_w9_031 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_031,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_031 );

dotpro_w9_032 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_032,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_032 );

dotpro_w9_033 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_033,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_033 );

dotpro_w9_034 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_034,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_034 );

dotpro_w9_035 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_035,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_035 );

dotpro_w9_036 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_036,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_036 );

dotpro_w9_037 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_037,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_037 );

dotpro_w9_038 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_038,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_038 );

dotpro_w9_039 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_039,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_039 );

dotpro_w9_040 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_040,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_040 );

dotpro_w9_041 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_041,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_041 );

dotpro_w9_042 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_042,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_042 );

dotpro_w9_043 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_043,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_043 );

dotpro_w9_044 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_044,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_044 );

dotpro_w9_045 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_045,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_045 );

dotpro_w9_046 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_046,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_046 );

dotpro_w9_047 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_047,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_047 );

dotpro_w9_048 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_048,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_048 );

dotpro_w9_049 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_049,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_049 );

dotpro_w9_050 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_050,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_050 );

dotpro_w9_051 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_051,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_051 );

dotpro_w9_052 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_052,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_052 );

dotpro_w9_053 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_053,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_053 );

dotpro_w9_054 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_054,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_054 );

dotpro_w9_055 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_055,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_055 );

dotpro_w9_056 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_056,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_056 );

dotpro_w9_057 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_057,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_057 );

dotpro_w9_058 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_058,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_058 );

dotpro_w9_059 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_059,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_059 );

dotpro_w9_060 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_060,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_060 );

dotpro_w9_061 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_061,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_061 );

dotpro_w9_062 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_062,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_062 );

dotpro_w9_063 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_063,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_063 );

dotpro_w9_064 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_064,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_064 );

dotpro_w9_065 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_065,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_065 );

dotpro_w9_066 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_066,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_066 );

dotpro_w9_067 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_067,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_067 );

dotpro_w9_068 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_068,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_068 );

dotpro_w9_069 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_069,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_069 );

dotpro_w9_070 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_070,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_070 );

dotpro_w9_071 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_071,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_071 );

dotpro_w9_072 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_072,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_072 );

dotpro_w9_073 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_073,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_073 );

dotpro_w9_074 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_074,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_074 );

dotpro_w9_075 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_075,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_075 );

dotpro_w9_076 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_076,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_076 );

dotpro_w9_077 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_077,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_077 );

dotpro_w9_078 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_078,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_078 );

dotpro_w9_079 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_079,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_079 );

dotpro_w9_080 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_080,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_080 );

dotpro_w9_081 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_081,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_081 );

dotpro_w9_082 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_082,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_082 );

dotpro_w9_083 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_083,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_083 );

dotpro_w9_084 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_084,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_084 );

dotpro_w9_085 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_085,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_085 );

dotpro_w9_086 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_086,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_086 );

dotpro_w9_087 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_087,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_087 );

dotpro_w9_088 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_088,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_088 );

dotpro_w9_089 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_089,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_089 );

dotpro_w9_090 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_090,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_090 );

dotpro_w9_091 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_091,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_091 );

dotpro_w9_092 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_092,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_092 );

dotpro_w9_093 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_093,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_093 );

dotpro_w9_094 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_094,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_094 );

dotpro_w9_095 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_095,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_095 );

dotpro_w9_096 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_096,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_096 );

dotpro_w9_097 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_097,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_097 );

dotpro_w9_098 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_098,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_098 );

dotpro_w9_099 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_099,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_099 );

dotpro_w9_100 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_100,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_100 );

dotpro_w9_101 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_101,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_101 );

dotpro_w9_102 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_102,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_102 );

dotpro_w9_103 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_103,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_103 );

dotpro_w9_104 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_104,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_104 );

dotpro_w9_105 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_105,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_105 );

dotpro_w9_106 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_106,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_106 );

dotpro_w9_107 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_107,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_107 );

dotpro_w9_108 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_108,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_108 );

dotpro_w9_109 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_109,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_109 );

dotpro_w9_110 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_110,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_110 );

dotpro_w9_111 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_111,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_111 );

dotpro_w9_112 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_112,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_112 );

dotpro_w9_113 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_113,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_113 );

dotpro_w9_114 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_114,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_114 );

dotpro_w9_115 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_115,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_115 );

dotpro_w9_116 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_116,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_116 );

dotpro_w9_117 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_117,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_117 );

dotpro_w9_118 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_118,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_118 );

dotpro_w9_119 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_119,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_119 );

dotpro_w9_120 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_120,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_120 );

dotpro_w9_121 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_121,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_121 );

dotpro_w9_122 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_122,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_122 );

dotpro_w9_123 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_123,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_123 );

dotpro_w9_124 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_124,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_124 );

dotpro_w9_125 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_125,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_125 );

dotpro_w9_126 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_126,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_126 );

dotpro_w9_127 : dotpro_w9 port map (
  CLK => CLK,
  RST_N => RST_N,
  PixRomData => PixRomData,
  MatRomData => coeff_127,
  CO9is1 => CO9is1,
  CO512is0 => CO512is0,
  MATMULOUT => mac_127 );

-------------------------------------------------------
  truncate_000 <= mac_000(13 downto 7);
  truncate_001 <= mac_001(13 downto 7);
  truncate_002 <= mac_002(13 downto 7);
  truncate_003 <= mac_003(13 downto 7);
  truncate_004 <= mac_004(13 downto 7);
  truncate_005 <= mac_005(13 downto 7);
  truncate_006 <= mac_006(13 downto 7);
  truncate_007 <= mac_007(13 downto 7);
  truncate_008 <= mac_008(13 downto 7);
  truncate_009 <= mac_009(13 downto 7);
  truncate_010 <= mac_010(13 downto 7);
  truncate_011 <= mac_011(13 downto 7);
  truncate_012 <= mac_012(13 downto 7);
  truncate_013 <= mac_013(13 downto 7);
  truncate_014 <= mac_014(13 downto 7);
  truncate_015 <= mac_015(13 downto 7);
  truncate_016 <= mac_016(13 downto 7);
  truncate_017 <= mac_017(13 downto 7);
  truncate_018 <= mac_018(13 downto 7);
  truncate_019 <= mac_019(13 downto 7);
  truncate_020 <= mac_020(13 downto 7);
  truncate_021 <= mac_021(13 downto 7);
  truncate_022 <= mac_022(13 downto 7);
  truncate_023 <= mac_023(13 downto 7);
  truncate_024 <= mac_024(13 downto 7);
  truncate_025 <= mac_025(13 downto 7);
  truncate_026 <= mac_026(13 downto 7);
  truncate_027 <= mac_027(13 downto 7);
  truncate_028 <= mac_028(13 downto 7);
  truncate_029 <= mac_029(13 downto 7);
  truncate_030 <= mac_030(13 downto 7);
  truncate_031 <= mac_031(13 downto 7);
  truncate_032 <= mac_032(13 downto 7);
  truncate_033 <= mac_033(13 downto 7);
  truncate_034 <= mac_034(13 downto 7);
  truncate_035 <= mac_035(13 downto 7);
  truncate_036 <= mac_036(13 downto 7);
  truncate_037 <= mac_037(13 downto 7);
  truncate_038 <= mac_038(13 downto 7);
  truncate_039 <= mac_039(13 downto 7);
  truncate_040 <= mac_040(13 downto 7);
  truncate_041 <= mac_041(13 downto 7);
  truncate_042 <= mac_042(13 downto 7);
  truncate_043 <= mac_043(13 downto 7);
  truncate_044 <= mac_044(13 downto 7);
  truncate_045 <= mac_045(13 downto 7);
  truncate_046 <= mac_046(13 downto 7);
  truncate_047 <= mac_047(13 downto 7);
  truncate_048 <= mac_048(13 downto 7);
  truncate_049 <= mac_049(13 downto 7);
  truncate_050 <= mac_050(13 downto 7);
  truncate_051 <= mac_051(13 downto 7);
  truncate_052 <= mac_052(13 downto 7);
  truncate_053 <= mac_053(13 downto 7);
  truncate_054 <= mac_054(13 downto 7);
  truncate_055 <= mac_055(13 downto 7);
  truncate_056 <= mac_056(13 downto 7);
  truncate_057 <= mac_057(13 downto 7);
  truncate_058 <= mac_058(13 downto 7);
  truncate_059 <= mac_059(13 downto 7);
  truncate_060 <= mac_060(13 downto 7);
  truncate_061 <= mac_061(13 downto 7);
  truncate_062 <= mac_062(13 downto 7);
  truncate_063 <= mac_063(13 downto 7);
  truncate_064 <= mac_064(13 downto 7);
  truncate_065 <= mac_065(13 downto 7);
  truncate_066 <= mac_066(13 downto 7);
  truncate_067 <= mac_067(13 downto 7);
  truncate_068 <= mac_068(13 downto 7);
  truncate_069 <= mac_069(13 downto 7);
  truncate_070 <= mac_070(13 downto 7);
  truncate_071 <= mac_071(13 downto 7);
  truncate_072 <= mac_072(13 downto 7);
  truncate_073 <= mac_073(13 downto 7);
  truncate_074 <= mac_074(13 downto 7);
  truncate_075 <= mac_075(13 downto 7);
  truncate_076 <= mac_076(13 downto 7);
  truncate_077 <= mac_077(13 downto 7);
  truncate_078 <= mac_078(13 downto 7);
  truncate_079 <= mac_079(13 downto 7);
  truncate_080 <= mac_080(13 downto 7);
  truncate_081 <= mac_081(13 downto 7);
  truncate_082 <= mac_082(13 downto 7);
  truncate_083 <= mac_083(13 downto 7);
  truncate_084 <= mac_084(13 downto 7);
  truncate_085 <= mac_085(13 downto 7);
  truncate_086 <= mac_086(13 downto 7);
  truncate_087 <= mac_087(13 downto 7);
  truncate_088 <= mac_088(13 downto 7);
  truncate_089 <= mac_089(13 downto 7);
  truncate_090 <= mac_090(13 downto 7);
  truncate_091 <= mac_091(13 downto 7);
  truncate_092 <= mac_092(13 downto 7);
  truncate_093 <= mac_093(13 downto 7);
  truncate_094 <= mac_094(13 downto 7);
  truncate_095 <= mac_095(13 downto 7);
  truncate_096 <= mac_096(13 downto 7);
  truncate_097 <= mac_097(13 downto 7);
  truncate_098 <= mac_098(13 downto 7);
  truncate_099 <= mac_099(13 downto 7);
  truncate_100 <= mac_100(13 downto 7);
  truncate_101 <= mac_101(13 downto 7);
  truncate_102 <= mac_102(13 downto 7);
  truncate_103 <= mac_103(13 downto 7);
  truncate_104 <= mac_104(13 downto 7);
  truncate_105 <= mac_105(13 downto 7);
  truncate_106 <= mac_106(13 downto 7);
  truncate_107 <= mac_107(13 downto 7);
  truncate_108 <= mac_108(13 downto 7);
  truncate_109 <= mac_109(13 downto 7);
  truncate_110 <= mac_110(13 downto 7);
  truncate_111 <= mac_111(13 downto 7);
  truncate_112 <= mac_112(13 downto 7);
  truncate_113 <= mac_113(13 downto 7);
  truncate_114 <= mac_114(13 downto 7);
  truncate_115 <= mac_115(13 downto 7);
  truncate_116 <= mac_116(13 downto 7);
  truncate_117 <= mac_117(13 downto 7);
  truncate_118 <= mac_118(13 downto 7);
  truncate_119 <= mac_119(13 downto 7);
  truncate_120 <= mac_120(13 downto 7);
  truncate_121 <= mac_121(13 downto 7);
  truncate_122 <= mac_122(13 downto 7);
  truncate_123 <= mac_123(13 downto 7);
  truncate_124 <= mac_124(13 downto 7);
  truncate_125 <= mac_125(13 downto 7);
  truncate_126 <= mac_126(13 downto 7);
  truncate_127 <= mac_127(13 downto 7);

---------------------------------------------------------
  Relu_000 <= (others => '0') when truncate_000 < 0 else
             (others => '1') when truncate_000 > 31 else
             truncate_000(4 downto 0);

  Relu_001 <= (others => '0') when truncate_001 < 0 else
             (others => '1') when truncate_001 > 31 else
             truncate_001(4 downto 0);

  Relu_002 <= (others => '0') when truncate_002 < 0 else
             (others => '1') when truncate_002 > 31 else
             truncate_002(4 downto 0);

  Relu_003 <= (others => '0') when truncate_003 < 0 else
             (others => '1') when truncate_003 > 31 else
             truncate_003(4 downto 0);

  Relu_004 <= (others => '0') when truncate_004 < 0 else
             (others => '1') when truncate_004 > 31 else
             truncate_004(4 downto 0);

  Relu_005 <= (others => '0') when truncate_005 < 0 else
             (others => '1') when truncate_005 > 31 else
             truncate_005(4 downto 0);

  Relu_006 <= (others => '0') when truncate_006 < 0 else
             (others => '1') when truncate_006 > 31 else
             truncate_006(4 downto 0);

  Relu_007 <= (others => '0') when truncate_007 < 0 else
             (others => '1') when truncate_007 > 31 else
             truncate_007(4 downto 0);

  Relu_008 <= (others => '0') when truncate_008 < 0 else
             (others => '1') when truncate_008 > 31 else
             truncate_008(4 downto 0);

  Relu_009 <= (others => '0') when truncate_009 < 0 else
             (others => '1') when truncate_009 > 31 else
             truncate_009(4 downto 0);

  Relu_010 <= (others => '0') when truncate_010 < 0 else
             (others => '1') when truncate_010 > 31 else
             truncate_010(4 downto 0);

  Relu_011 <= (others => '0') when truncate_011 < 0 else
             (others => '1') when truncate_011 > 31 else
             truncate_011(4 downto 0);

  Relu_012 <= (others => '0') when truncate_012 < 0 else
             (others => '1') when truncate_012 > 31 else
             truncate_012(4 downto 0);

  Relu_013 <= (others => '0') when truncate_013 < 0 else
             (others => '1') when truncate_013 > 31 else
             truncate_013(4 downto 0);

  Relu_014 <= (others => '0') when truncate_014 < 0 else
             (others => '1') when truncate_014 > 31 else
             truncate_014(4 downto 0);

  Relu_015 <= (others => '0') when truncate_015 < 0 else
             (others => '1') when truncate_015 > 31 else
             truncate_015(4 downto 0);

  Relu_016 <= (others => '0') when truncate_016 < 0 else
             (others => '1') when truncate_016 > 31 else
             truncate_016(4 downto 0);

  Relu_017 <= (others => '0') when truncate_017 < 0 else
             (others => '1') when truncate_017 > 31 else
             truncate_017(4 downto 0);

  Relu_018 <= (others => '0') when truncate_018 < 0 else
             (others => '1') when truncate_018 > 31 else
             truncate_018(4 downto 0);

  Relu_019 <= (others => '0') when truncate_019 < 0 else
             (others => '1') when truncate_019 > 31 else
             truncate_019(4 downto 0);

  Relu_020 <= (others => '0') when truncate_020 < 0 else
             (others => '1') when truncate_020 > 31 else
             truncate_020(4 downto 0);

  Relu_021 <= (others => '0') when truncate_021 < 0 else
             (others => '1') when truncate_021 > 31 else
             truncate_021(4 downto 0);

  Relu_022 <= (others => '0') when truncate_022 < 0 else
             (others => '1') when truncate_022 > 31 else
             truncate_022(4 downto 0);

  Relu_023 <= (others => '0') when truncate_023 < 0 else
             (others => '1') when truncate_023 > 31 else
             truncate_023(4 downto 0);

  Relu_024 <= (others => '0') when truncate_024 < 0 else
             (others => '1') when truncate_024 > 31 else
             truncate_024(4 downto 0);

  Relu_025 <= (others => '0') when truncate_025 < 0 else
             (others => '1') when truncate_025 > 31 else
             truncate_025(4 downto 0);

  Relu_026 <= (others => '0') when truncate_026 < 0 else
             (others => '1') when truncate_026 > 31 else
             truncate_026(4 downto 0);

  Relu_027 <= (others => '0') when truncate_027 < 0 else
             (others => '1') when truncate_027 > 31 else
             truncate_027(4 downto 0);

  Relu_028 <= (others => '0') when truncate_028 < 0 else
             (others => '1') when truncate_028 > 31 else
             truncate_028(4 downto 0);

  Relu_029 <= (others => '0') when truncate_029 < 0 else
             (others => '1') when truncate_029 > 31 else
             truncate_029(4 downto 0);

  Relu_030 <= (others => '0') when truncate_030 < 0 else
             (others => '1') when truncate_030 > 31 else
             truncate_030(4 downto 0);

  Relu_031 <= (others => '0') when truncate_031 < 0 else
             (others => '1') when truncate_031 > 31 else
             truncate_031(4 downto 0);

  Relu_032 <= (others => '0') when truncate_032 < 0 else
             (others => '1') when truncate_032 > 31 else
             truncate_032(4 downto 0);

  Relu_033 <= (others => '0') when truncate_033 < 0 else
             (others => '1') when truncate_033 > 31 else
             truncate_033(4 downto 0);

  Relu_034 <= (others => '0') when truncate_034 < 0 else
             (others => '1') when truncate_034 > 31 else
             truncate_034(4 downto 0);

  Relu_035 <= (others => '0') when truncate_035 < 0 else
             (others => '1') when truncate_035 > 31 else
             truncate_035(4 downto 0);

  Relu_036 <= (others => '0') when truncate_036 < 0 else
             (others => '1') when truncate_036 > 31 else
             truncate_036(4 downto 0);

  Relu_037 <= (others => '0') when truncate_037 < 0 else
             (others => '1') when truncate_037 > 31 else
             truncate_037(4 downto 0);

  Relu_038 <= (others => '0') when truncate_038 < 0 else
             (others => '1') when truncate_038 > 31 else
             truncate_038(4 downto 0);

  Relu_039 <= (others => '0') when truncate_039 < 0 else
             (others => '1') when truncate_039 > 31 else
             truncate_039(4 downto 0);

  Relu_040 <= (others => '0') when truncate_040 < 0 else
             (others => '1') when truncate_040 > 31 else
             truncate_040(4 downto 0);

  Relu_041 <= (others => '0') when truncate_041 < 0 else
             (others => '1') when truncate_041 > 31 else
             truncate_041(4 downto 0);

  Relu_042 <= (others => '0') when truncate_042 < 0 else
             (others => '1') when truncate_042 > 31 else
             truncate_042(4 downto 0);

  Relu_043 <= (others => '0') when truncate_043 < 0 else
             (others => '1') when truncate_043 > 31 else
             truncate_043(4 downto 0);

  Relu_044 <= (others => '0') when truncate_044 < 0 else
             (others => '1') when truncate_044 > 31 else
             truncate_044(4 downto 0);

  Relu_045 <= (others => '0') when truncate_045 < 0 else
             (others => '1') when truncate_045 > 31 else
             truncate_045(4 downto 0);

  Relu_046 <= (others => '0') when truncate_046 < 0 else
             (others => '1') when truncate_046 > 31 else
             truncate_046(4 downto 0);

  Relu_047 <= (others => '0') when truncate_047 < 0 else
             (others => '1') when truncate_047 > 31 else
             truncate_047(4 downto 0);

  Relu_048 <= (others => '0') when truncate_048 < 0 else
             (others => '1') when truncate_048 > 31 else
             truncate_048(4 downto 0);

  Relu_049 <= (others => '0') when truncate_049 < 0 else
             (others => '1') when truncate_049 > 31 else
             truncate_049(4 downto 0);

  Relu_050 <= (others => '0') when truncate_050 < 0 else
             (others => '1') when truncate_050 > 31 else
             truncate_050(4 downto 0);

  Relu_051 <= (others => '0') when truncate_051 < 0 else
             (others => '1') when truncate_051 > 31 else
             truncate_051(4 downto 0);

  Relu_052 <= (others => '0') when truncate_052 < 0 else
             (others => '1') when truncate_052 > 31 else
             truncate_052(4 downto 0);

  Relu_053 <= (others => '0') when truncate_053 < 0 else
             (others => '1') when truncate_053 > 31 else
             truncate_053(4 downto 0);

  Relu_054 <= (others => '0') when truncate_054 < 0 else
             (others => '1') when truncate_054 > 31 else
             truncate_054(4 downto 0);

  Relu_055 <= (others => '0') when truncate_055 < 0 else
             (others => '1') when truncate_055 > 31 else
             truncate_055(4 downto 0);

  Relu_056 <= (others => '0') when truncate_056 < 0 else
             (others => '1') when truncate_056 > 31 else
             truncate_056(4 downto 0);

  Relu_057 <= (others => '0') when truncate_057 < 0 else
             (others => '1') when truncate_057 > 31 else
             truncate_057(4 downto 0);

  Relu_058 <= (others => '0') when truncate_058 < 0 else
             (others => '1') when truncate_058 > 31 else
             truncate_058(4 downto 0);

  Relu_059 <= (others => '0') when truncate_059 < 0 else
             (others => '1') when truncate_059 > 31 else
             truncate_059(4 downto 0);

  Relu_060 <= (others => '0') when truncate_060 < 0 else
             (others => '1') when truncate_060 > 31 else
             truncate_060(4 downto 0);

  Relu_061 <= (others => '0') when truncate_061 < 0 else
             (others => '1') when truncate_061 > 31 else
             truncate_061(4 downto 0);

  Relu_062 <= (others => '0') when truncate_062 < 0 else
             (others => '1') when truncate_062 > 31 else
             truncate_062(4 downto 0);

  Relu_063 <= (others => '0') when truncate_063 < 0 else
             (others => '1') when truncate_063 > 31 else
             truncate_063(4 downto 0);

  Relu_064 <= (others => '0') when truncate_064 < 0 else
             (others => '1') when truncate_064 > 31 else
             truncate_064(4 downto 0);

  Relu_065 <= (others => '0') when truncate_065 < 0 else
             (others => '1') when truncate_065 > 31 else
             truncate_065(4 downto 0);

  Relu_066 <= (others => '0') when truncate_066 < 0 else
             (others => '1') when truncate_066 > 31 else
             truncate_066(4 downto 0);

  Relu_067 <= (others => '0') when truncate_067 < 0 else
             (others => '1') when truncate_067 > 31 else
             truncate_067(4 downto 0);

  Relu_068 <= (others => '0') when truncate_068 < 0 else
             (others => '1') when truncate_068 > 31 else
             truncate_068(4 downto 0);

  Relu_069 <= (others => '0') when truncate_069 < 0 else
             (others => '1') when truncate_069 > 31 else
             truncate_069(4 downto 0);

  Relu_070 <= (others => '0') when truncate_070 < 0 else
             (others => '1') when truncate_070 > 31 else
             truncate_070(4 downto 0);

  Relu_071 <= (others => '0') when truncate_071 < 0 else
             (others => '1') when truncate_071 > 31 else
             truncate_071(4 downto 0);

  Relu_072 <= (others => '0') when truncate_072 < 0 else
             (others => '1') when truncate_072 > 31 else
             truncate_072(4 downto 0);

  Relu_073 <= (others => '0') when truncate_073 < 0 else
             (others => '1') when truncate_073 > 31 else
             truncate_073(4 downto 0);

  Relu_074 <= (others => '0') when truncate_074 < 0 else
             (others => '1') when truncate_074 > 31 else
             truncate_074(4 downto 0);

  Relu_075 <= (others => '0') when truncate_075 < 0 else
             (others => '1') when truncate_075 > 31 else
             truncate_075(4 downto 0);

  Relu_076 <= (others => '0') when truncate_076 < 0 else
             (others => '1') when truncate_076 > 31 else
             truncate_076(4 downto 0);

  Relu_077 <= (others => '0') when truncate_077 < 0 else
             (others => '1') when truncate_077 > 31 else
             truncate_077(4 downto 0);

  Relu_078 <= (others => '0') when truncate_078 < 0 else
             (others => '1') when truncate_078 > 31 else
             truncate_078(4 downto 0);

  Relu_079 <= (others => '0') when truncate_079 < 0 else
             (others => '1') when truncate_079 > 31 else
             truncate_079(4 downto 0);

  Relu_080 <= (others => '0') when truncate_080 < 0 else
             (others => '1') when truncate_080 > 31 else
             truncate_080(4 downto 0);

  Relu_081 <= (others => '0') when truncate_081 < 0 else
             (others => '1') when truncate_081 > 31 else
             truncate_081(4 downto 0);

  Relu_082 <= (others => '0') when truncate_082 < 0 else
             (others => '1') when truncate_082 > 31 else
             truncate_082(4 downto 0);

  Relu_083 <= (others => '0') when truncate_083 < 0 else
             (others => '1') when truncate_083 > 31 else
             truncate_083(4 downto 0);

  Relu_084 <= (others => '0') when truncate_084 < 0 else
             (others => '1') when truncate_084 > 31 else
             truncate_084(4 downto 0);

  Relu_085 <= (others => '0') when truncate_085 < 0 else
             (others => '1') when truncate_085 > 31 else
             truncate_085(4 downto 0);

  Relu_086 <= (others => '0') when truncate_086 < 0 else
             (others => '1') when truncate_086 > 31 else
             truncate_086(4 downto 0);

  Relu_087 <= (others => '0') when truncate_087 < 0 else
             (others => '1') when truncate_087 > 31 else
             truncate_087(4 downto 0);

  Relu_088 <= (others => '0') when truncate_088 < 0 else
             (others => '1') when truncate_088 > 31 else
             truncate_088(4 downto 0);

  Relu_089 <= (others => '0') when truncate_089 < 0 else
             (others => '1') when truncate_089 > 31 else
             truncate_089(4 downto 0);

  Relu_090 <= (others => '0') when truncate_090 < 0 else
             (others => '1') when truncate_090 > 31 else
             truncate_090(4 downto 0);

  Relu_091 <= (others => '0') when truncate_091 < 0 else
             (others => '1') when truncate_091 > 31 else
             truncate_091(4 downto 0);

  Relu_092 <= (others => '0') when truncate_092 < 0 else
             (others => '1') when truncate_092 > 31 else
             truncate_092(4 downto 0);

  Relu_093 <= (others => '0') when truncate_093 < 0 else
             (others => '1') when truncate_093 > 31 else
             truncate_093(4 downto 0);

  Relu_094 <= (others => '0') when truncate_094 < 0 else
             (others => '1') when truncate_094 > 31 else
             truncate_094(4 downto 0);

  Relu_095 <= (others => '0') when truncate_095 < 0 else
             (others => '1') when truncate_095 > 31 else
             truncate_095(4 downto 0);

  Relu_096 <= (others => '0') when truncate_096 < 0 else
             (others => '1') when truncate_096 > 31 else
             truncate_096(4 downto 0);

  Relu_097 <= (others => '0') when truncate_097 < 0 else
             (others => '1') when truncate_097 > 31 else
             truncate_097(4 downto 0);

  Relu_098 <= (others => '0') when truncate_098 < 0 else
             (others => '1') when truncate_098 > 31 else
             truncate_098(4 downto 0);

  Relu_099 <= (others => '0') when truncate_099 < 0 else
             (others => '1') when truncate_099 > 31 else
             truncate_099(4 downto 0);

  Relu_100 <= (others => '0') when truncate_100 < 0 else
             (others => '1') when truncate_100 > 31 else
             truncate_100(4 downto 0);

  Relu_101 <= (others => '0') when truncate_101 < 0 else
             (others => '1') when truncate_101 > 31 else
             truncate_101(4 downto 0);

  Relu_102 <= (others => '0') when truncate_102 < 0 else
             (others => '1') when truncate_102 > 31 else
             truncate_102(4 downto 0);

  Relu_103 <= (others => '0') when truncate_103 < 0 else
             (others => '1') when truncate_103 > 31 else
             truncate_103(4 downto 0);

  Relu_104 <= (others => '0') when truncate_104 < 0 else
             (others => '1') when truncate_104 > 31 else
             truncate_104(4 downto 0);

  Relu_105 <= (others => '0') when truncate_105 < 0 else
             (others => '1') when truncate_105 > 31 else
             truncate_105(4 downto 0);

  Relu_106 <= (others => '0') when truncate_106 < 0 else
             (others => '1') when truncate_106 > 31 else
             truncate_106(4 downto 0);

  Relu_107 <= (others => '0') when truncate_107 < 0 else
             (others => '1') when truncate_107 > 31 else
             truncate_107(4 downto 0);

  Relu_108 <= (others => '0') when truncate_108 < 0 else
             (others => '1') when truncate_108 > 31 else
             truncate_108(4 downto 0);

  Relu_109 <= (others => '0') when truncate_109 < 0 else
             (others => '1') when truncate_109 > 31 else
             truncate_109(4 downto 0);

  Relu_110 <= (others => '0') when truncate_110 < 0 else
             (others => '1') when truncate_110 > 31 else
             truncate_110(4 downto 0);

  Relu_111 <= (others => '0') when truncate_111 < 0 else
             (others => '1') when truncate_111 > 31 else
             truncate_111(4 downto 0);

  Relu_112 <= (others => '0') when truncate_112 < 0 else
             (others => '1') when truncate_112 > 31 else
             truncate_112(4 downto 0);

  Relu_113 <= (others => '0') when truncate_113 < 0 else
             (others => '1') when truncate_113 > 31 else
             truncate_113(4 downto 0);

  Relu_114 <= (others => '0') when truncate_114 < 0 else
             (others => '1') when truncate_114 > 31 else
             truncate_114(4 downto 0);

  Relu_115 <= (others => '0') when truncate_115 < 0 else
             (others => '1') when truncate_115 > 31 else
             truncate_115(4 downto 0);

  Relu_116 <= (others => '0') when truncate_116 < 0 else
             (others => '1') when truncate_116 > 31 else
             truncate_116(4 downto 0);

  Relu_117 <= (others => '0') when truncate_117 < 0 else
             (others => '1') when truncate_117 > 31 else
             truncate_117(4 downto 0);

  Relu_118 <= (others => '0') when truncate_118 < 0 else
             (others => '1') when truncate_118 > 31 else
             truncate_118(4 downto 0);

  Relu_119 <= (others => '0') when truncate_119 < 0 else
             (others => '1') when truncate_119 > 31 else
             truncate_119(4 downto 0);

  Relu_120 <= (others => '0') when truncate_120 < 0 else
             (others => '1') when truncate_120 > 31 else
             truncate_120(4 downto 0);

  Relu_121 <= (others => '0') when truncate_121 < 0 else
             (others => '1') when truncate_121 > 31 else
             truncate_121(4 downto 0);

  Relu_122 <= (others => '0') when truncate_122 < 0 else
             (others => '1') when truncate_122 > 31 else
             truncate_122(4 downto 0);

  Relu_123 <= (others => '0') when truncate_123 < 0 else
             (others => '1') when truncate_123 > 31 else
             truncate_123(4 downto 0);

  Relu_124 <= (others => '0') when truncate_124 < 0 else
             (others => '1') when truncate_124 > 31 else
             truncate_124(4 downto 0);

  Relu_125 <= (others => '0') when truncate_125 < 0 else
             (others => '1') when truncate_125 > 31 else
             truncate_125(4 downto 0);

  Relu_126 <= (others => '0') when truncate_126 < 0 else
             (others => '1') when truncate_126 > 31 else
             truncate_126(4 downto 0);

  Relu_127 <= (others => '0') when truncate_127 < 0 else
             (others => '1') when truncate_127 > 31 else
             truncate_127(4 downto 0);


--ReluSum <= ("0000000" & Relu_00) + ("0000000" & Relu_01)
--		+ ("0000000" & Relu_02) + ("0000000" & Relu_03)
--		+ ("0000000" & Relu_04) + ("0000000" & Relu_05)
--		+ ("0000000" & Relu_06) + ("0000000" & Relu_07)
--		+ ("0000000" & Relu_08) + ("0000000" & Relu_09)
--
--		+ ("0000000" & Relu_10) + ("0000000" & Relu_11)
--		+ ("0000000" & Relu_12) + ("0000000" & Relu_13)
--		+ ("0000000" & Relu_14) + ("0000000" & Relu_15)
--		+ ("0000000" & Relu_16) + ("0000000" & Relu_17)
--		+ ("0000000" & Relu_18) + ("0000000" & Relu_19)
--
--		+ ("0000000" & Relu_20) + ("0000000" & Relu_21)
--		+ ("0000000" & Relu_22) + ("0000000" & Relu_23)
--		+ ("0000000" & Relu_24) + ("0000000" & Relu_25)
--		+ ("0000000" & Relu_26) + ("0000000" & Relu_27)
--		+ ("0000000" & Relu_28) + ("0000000" & Relu_29)
--
--		+ ("0000000" & Relu_30) + ("0000000" & Relu_31);

ReluAll <= 
	reg_000_0 & reg_000_1 & reg_000_2 & reg_000_3 & 
	reg_001_0 & reg_001_1 & reg_001_2 & reg_001_3 & 
	reg_002_0 & reg_002_1 & reg_002_2 & reg_002_3 & 
	reg_003_0 & reg_003_1 & reg_003_2 & reg_003_3 & 
	reg_004_0 & reg_004_1 & reg_004_2 & reg_004_3 & 
	reg_005_0 & reg_005_1 & reg_005_2 & reg_005_3 & 
	reg_006_0 & reg_006_1 & reg_006_2 & reg_006_3 & 
	reg_007_0 & reg_007_1 & reg_007_2 & reg_007_3 & 
	reg_008_0 & reg_008_1 & reg_008_2 & reg_008_3 & 
	reg_009_0 & reg_009_1 & reg_009_2 & reg_009_3 & 
	reg_010_0 & reg_010_1 & reg_010_2 & reg_010_3 & 
	reg_011_0 & reg_011_1 & reg_011_2 & reg_011_3 & 
	reg_012_0 & reg_012_1 & reg_012_2 & reg_012_3 & 
	reg_013_0 & reg_013_1 & reg_013_2 & reg_013_3 & 
	reg_014_0 & reg_014_1 & reg_014_2 & reg_014_3 & 
	reg_015_0 & reg_015_1 & reg_015_2 & reg_015_3 & 
	reg_016_0 & reg_016_1 & reg_016_2 & reg_016_3 & 
	reg_017_0 & reg_017_1 & reg_017_2 & reg_017_3 & 
	reg_018_0 & reg_018_1 & reg_018_2 & reg_018_3 & 
	reg_019_0 & reg_019_1 & reg_019_2 & reg_019_3 & 
	reg_020_0 & reg_020_1 & reg_020_2 & reg_020_3 & 
	reg_021_0 & reg_021_1 & reg_021_2 & reg_021_3 & 
	reg_022_0 & reg_022_1 & reg_022_2 & reg_022_3 & 
	reg_023_0 & reg_023_1 & reg_023_2 & reg_023_3 & 
	reg_024_0 & reg_024_1 & reg_024_2 & reg_024_3 & 
	reg_025_0 & reg_025_1 & reg_025_2 & reg_025_3 & 
	reg_026_0 & reg_026_1 & reg_026_2 & reg_026_3 & 
	reg_027_0 & reg_027_1 & reg_027_2 & reg_027_3 & 
	reg_028_0 & reg_028_1 & reg_028_2 & reg_028_3 & 
	reg_029_0 & reg_029_1 & reg_029_2 & reg_029_3 & 
	reg_030_0 & reg_030_1 & reg_030_2 & reg_030_3 & 
	reg_031_0 & reg_031_1 & reg_031_2 & reg_031_3 & 
	reg_032_0 & reg_032_1 & reg_032_2 & reg_032_3 & 
	reg_033_0 & reg_033_1 & reg_033_2 & reg_033_3 & 
	reg_034_0 & reg_034_1 & reg_034_2 & reg_034_3 & 
	reg_035_0 & reg_035_1 & reg_035_2 & reg_035_3 & 
	reg_036_0 & reg_036_1 & reg_036_2 & reg_036_3 & 
	reg_037_0 & reg_037_1 & reg_037_2 & reg_037_3 & 
	reg_038_0 & reg_038_1 & reg_038_2 & reg_038_3 & 
	reg_039_0 & reg_039_1 & reg_039_2 & reg_039_3 & 
	reg_040_0 & reg_040_1 & reg_040_2 & reg_040_3 & 
	reg_041_0 & reg_041_1 & reg_041_2 & reg_041_3 & 
	reg_042_0 & reg_042_1 & reg_042_2 & reg_042_3 & 
	reg_043_0 & reg_043_1 & reg_043_2 & reg_043_3 & 
	reg_044_0 & reg_044_1 & reg_044_2 & reg_044_3 & 
	reg_045_0 & reg_045_1 & reg_045_2 & reg_045_3 & 
	reg_046_0 & reg_046_1 & reg_046_2 & reg_046_3 & 
	reg_047_0 & reg_047_1 & reg_047_2 & reg_047_3 & 
	reg_048_0 & reg_048_1 & reg_048_2 & reg_048_3 & 
	reg_049_0 & reg_049_1 & reg_049_2 & reg_049_3 & 
	reg_050_0 & reg_050_1 & reg_050_2 & reg_050_3 & 
	reg_051_0 & reg_051_1 & reg_051_2 & reg_051_3 & 
	reg_052_0 & reg_052_1 & reg_052_2 & reg_052_3 & 
	reg_053_0 & reg_053_1 & reg_053_2 & reg_053_3 & 
	reg_054_0 & reg_054_1 & reg_054_2 & reg_054_3 & 
	reg_055_0 & reg_055_1 & reg_055_2 & reg_055_3 & 
	reg_056_0 & reg_056_1 & reg_056_2 & reg_056_3 & 
	reg_057_0 & reg_057_1 & reg_057_2 & reg_057_3 & 
	reg_058_0 & reg_058_1 & reg_058_2 & reg_058_3 & 
	reg_059_0 & reg_059_1 & reg_059_2 & reg_059_3 & 
	reg_060_0 & reg_060_1 & reg_060_2 & reg_060_3 & 
	reg_061_0 & reg_061_1 & reg_061_2 & reg_061_3 & 
	reg_062_0 & reg_062_1 & reg_062_2 & reg_062_3 & 
	reg_063_0 & reg_063_1 & reg_063_2 & reg_063_3 & 
	reg_064_0 & reg_064_1 & reg_064_2 & reg_064_3 & 
	reg_065_0 & reg_065_1 & reg_065_2 & reg_065_3 & 
	reg_066_0 & reg_066_1 & reg_066_2 & reg_066_3 & 
	reg_067_0 & reg_067_1 & reg_067_2 & reg_067_3 & 
	reg_068_0 & reg_068_1 & reg_068_2 & reg_068_3 & 
	reg_069_0 & reg_069_1 & reg_069_2 & reg_069_3 & 
	reg_070_0 & reg_070_1 & reg_070_2 & reg_070_3 & 
	reg_071_0 & reg_071_1 & reg_071_2 & reg_071_3 & 
	reg_072_0 & reg_072_1 & reg_072_2 & reg_072_3 & 
	reg_073_0 & reg_073_1 & reg_073_2 & reg_073_3 & 
	reg_074_0 & reg_074_1 & reg_074_2 & reg_074_3 & 
	reg_075_0 & reg_075_1 & reg_075_2 & reg_075_3 & 
	reg_076_0 & reg_076_1 & reg_076_2 & reg_076_3 & 
	reg_077_0 & reg_077_1 & reg_077_2 & reg_077_3 & 
	reg_078_0 & reg_078_1 & reg_078_2 & reg_078_3 & 
	reg_079_0 & reg_079_1 & reg_079_2 & reg_079_3 & 
	reg_080_0 & reg_080_1 & reg_080_2 & reg_080_3 & 
	reg_081_0 & reg_081_1 & reg_081_2 & reg_081_3 & 
	reg_082_0 & reg_082_1 & reg_082_2 & reg_082_3 & 
	reg_083_0 & reg_083_1 & reg_083_2 & reg_083_3 & 
	reg_084_0 & reg_084_1 & reg_084_2 & reg_084_3 & 
	reg_085_0 & reg_085_1 & reg_085_2 & reg_085_3 & 
	reg_086_0 & reg_086_1 & reg_086_2 & reg_086_3 & 
	reg_087_0 & reg_087_1 & reg_087_2 & reg_087_3 & 
	reg_088_0 & reg_088_1 & reg_088_2 & reg_088_3 & 
	reg_089_0 & reg_089_1 & reg_089_2 & reg_089_3 & 
	reg_090_0 & reg_090_1 & reg_090_2 & reg_090_3 & 
	reg_091_0 & reg_091_1 & reg_091_2 & reg_091_3 & 
	reg_092_0 & reg_092_1 & reg_092_2 & reg_092_3 & 
	reg_093_0 & reg_093_1 & reg_093_2 & reg_093_3 & 
	reg_094_0 & reg_094_1 & reg_094_2 & reg_094_3 & 
	reg_095_0 & reg_095_1 & reg_095_2 & reg_095_3 & 
	reg_096_0 & reg_096_1 & reg_096_2 & reg_096_3 & 
	reg_097_0 & reg_097_1 & reg_097_2 & reg_097_3 & 
	reg_098_0 & reg_098_1 & reg_098_2 & reg_098_3 & 
	reg_099_0 & reg_099_1 & reg_099_2 & reg_099_3 & 
	reg_100_0 & reg_100_1 & reg_100_2 & reg_100_3 & 
	reg_101_0 & reg_101_1 & reg_101_2 & reg_101_3 & 
	reg_102_0 & reg_102_1 & reg_102_2 & reg_102_3 & 
	reg_103_0 & reg_103_1 & reg_103_2 & reg_103_3 & 
	reg_104_0 & reg_104_1 & reg_104_2 & reg_104_3 & 
	reg_105_0 & reg_105_1 & reg_105_2 & reg_105_3 & 
	reg_106_0 & reg_106_1 & reg_106_2 & reg_106_3 & 
	reg_107_0 & reg_107_1 & reg_107_2 & reg_107_3 & 
	reg_108_0 & reg_108_1 & reg_108_2 & reg_108_3 & 
	reg_109_0 & reg_109_1 & reg_109_2 & reg_109_3 & 
	reg_110_0 & reg_110_1 & reg_110_2 & reg_110_3 & 
	reg_111_0 & reg_111_1 & reg_111_2 & reg_111_3 & 
	reg_112_0 & reg_112_1 & reg_112_2 & reg_112_3 & 
	reg_113_0 & reg_113_1 & reg_113_2 & reg_113_3 & 
	reg_114_0 & reg_114_1 & reg_114_2 & reg_114_3 & 
	reg_115_0 & reg_115_1 & reg_115_2 & reg_115_3 & 
	reg_116_0 & reg_116_1 & reg_116_2 & reg_116_3 & 
	reg_117_0 & reg_117_1 & reg_117_2 & reg_117_3 & 
	reg_118_0 & reg_118_1 & reg_118_2 & reg_118_3 & 
	reg_119_0 & reg_119_1 & reg_119_2 & reg_119_3 & 
	reg_120_0 & reg_120_1 & reg_120_2 & reg_120_3 & 
	reg_121_0 & reg_121_1 & reg_121_2 & reg_121_3 & 
	reg_122_0 & reg_122_1 & reg_122_2 & reg_122_3 & 
	reg_123_0 & reg_123_1 & reg_123_2 & reg_123_3 & 
	reg_124_0 & reg_124_1 & reg_124_2 & reg_124_3 & 
	reg_125_0 & reg_125_1 & reg_125_2 & reg_125_3 & 
	reg_126_0 & reg_126_1 & reg_126_2 & reg_126_3 & 
	reg_127_0 & reg_127_1 & reg_127_2 & reg_127_3 ; 

ReluAllWrN <= not enb_a;

end RTL;

